"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.UBAClientWithRefresh = void 0;
var tslib_1 = require("tslib");
var assert_1 = tslib_1.__importDefault(require("assert"));
var UBAClientUtilities_1 = require("./UBAClientUtilities");
var interfaces_1 = require("../../interfaces");
var utils_1 = require("../../utils");
var UBAFeeCalculator_1 = require("../../UBAFeeCalculator");
var UBAFeeSpokeCalculatorAnalog_1 = require("../../UBAFeeCalculator/UBAFeeSpokeCalculatorAnalog");
var ethers_1 = require("ethers");
var BaseAbstractClient_1 = require("../BaseAbstractClient");
var UBAClientWithRefresh = (function (_super) {
    tslib_1.__extends(UBAClientWithRefresh, _super);
    function UBAClientWithRefresh(clientConfig, tokens, hubPoolClient, spokePoolClients, cachingClient) {
        var _this = _super.call(this, cachingClient) || this;
        _this.clientConfig = clientConfig;
        _this.tokens = tokens;
        _this.hubPoolClient = hubPoolClient;
        _this.spokePoolClients = spokePoolClients;
        _this.ubaClientState = {};
        _this.ubaBundleStates = {};
        _this.ubaBundleBlockRanges = [];
        _this.latestBlockTimestamps = {};
        _this.getKeyForBundle = function (bundleBlockRanges, tokenSymbol, chainId) {
            return "UBA_BUNDLE_STATE_".concat((0, UBAClientUtilities_1.getBundleKeyForBlockRanges)(bundleBlockRanges), "-").concat(tokenSymbol, "-").concat(chainId);
        };
        _this.logger = _this.hubPoolClient.logger;
        _this.enabledChainIds = _this.hubPoolClient.configStoreClient.getEnabledChains();
        (0, assert_1.default)(_this.enabledChainIds.length > 0, "No chainIds provided");
        _this.enabledChainIds.forEach(function (chainId) {
            (0, assert_1.default)((0, utils_1.isDefined)(spokePoolClients[chainId]), "No SpokePool provided for chainId ".concat(chainId));
        });
        _this.chainIdIndices = _this.hubPoolClient.configStoreClient.getChainIdIndicesForBlock();
        return _this;
    }
    UBAClientWithRefresh.prototype.computeFeesForDeposit = function (deposit) {
        var _a;
        this.assertUpdated();
        var tokenSymbol = (_a = this.hubPoolClient.getL1TokenInfoForL2Token(deposit.originToken, deposit.originChainId)) === null || _a === void 0 ? void 0 : _a.symbol;
        if (!tokenSymbol)
            throw new Error("No token symbol found");
        var blockRangeContainingDeposit = this.getUbaBundleBlockRangeContainingFlow(deposit.blockNumber, deposit.originChainId);
        var specificBundleState = this.getBundleState(blockRangeContainingDeposit, tokenSymbol, deposit.originChainId);
        var matchingFlow = (0, UBAClientUtilities_1.getMatchingFlow)(specificBundleState.flows, deposit);
        if (!matchingFlow) {
            throw new Error("Found bundle state containing flow but no matching flow found for deposit");
        }
        var lpFee = matchingFlow.lpFee.mul(deposit.amount).div(utils_1.fixedPointAdjustment);
        var depositBalancingFee = matchingFlow.balancingFee;
        return {
            lpFee: lpFee,
            depositBalancingFee: depositBalancingFee,
            systemFee: lpFee.add(depositBalancingFee),
        };
    };
    UBAClientWithRefresh.prototype.computeBalancingFeeForNextRefund = function (repaymentChainId, refundTokenSymbol, amount) {
        this.assertUpdated();
        var latestBlockRange = this.ubaBundleBlockRanges.at(-1);
        if (!latestBlockRange) {
            throw new Error("No block ranges stored, have you updated this client?");
        }
        var lastBundleState = this.getBundleState(latestBlockRange, refundTokenSymbol, repaymentChainId);
        var latestRunningBalance = ethers_1.ethers.constants.Zero;
        var latestIncentiveBalance = ethers_1.ethers.constants.Zero;
        var ubaConfigForBundle = lastBundleState.ubaConfig;
        var lastFlow = lastBundleState.flows.at(-1);
        if ((0, utils_1.isDefined)(lastFlow)) {
            var _a = UBAFeeCalculator_1.analog.calculateHistoricalRunningBalance(lastBundleState.flows.map(function (_a) {
                var flow = _a.flow, balancingFee = _a.balancingFee;
                return tslib_1.__assign(tslib_1.__assign({}, flow), { incentiveFee: balancingFee });
            }), lastFlow.runningBalance, lastFlow.incentiveBalance, lastFlow.netRunningBalanceAdjustment, repaymentChainId, refundTokenSymbol, ubaConfigForBundle), runningBalance = _a.runningBalance, incentiveBalance = _a.incentiveBalance;
            latestRunningBalance = runningBalance;
            latestIncentiveBalance = incentiveBalance;
        }
        var balancingFee = (0, UBAFeeSpokeCalculatorAnalog_1.getRefundFee)(amount, latestRunningBalance, latestIncentiveBalance, repaymentChainId, ubaConfigForBundle).balancingFee;
        return balancingFee;
    };
    UBAClientWithRefresh.prototype.getModifiedFlows = function (chainId, tokenSymbol, startBlock, endBlock) {
        var _this = this;
        this.assertUpdated();
        var specificBundleRange = this.ubaBundleBlockRanges.find(function (blockRanges) {
            var blockRangeForChain = (0, utils_1.getBlockRangeForChain)(blockRanges, chainId, _this.chainIdIndices);
            return blockRangeForChain[0] <= startBlock && blockRangeForChain[1] >= endBlock;
        });
        if (!(0, utils_1.isDefined)(specificBundleRange)) {
            throw new Error("Could not find bundle block range containing block range ".concat(startBlock, " to ").concat(endBlock, " on chain ").concat(chainId));
        }
        var bundleState = this.getBundleState(specificBundleRange, tokenSymbol, chainId);
        return bundleState.flows.filter(function (_a) {
            var flow = _a.flow;
            flow.blockNumber <= endBlock && flow.blockNumber >= startBlock;
        });
    };
    UBAClientWithRefresh.prototype.getLatestBundleState = function (tokenSymbol, chainId) {
        this.assertUpdated();
        var latestBlockRange = this.ubaBundleBlockRanges.at(-1);
        if (!latestBlockRange) {
            throw new Error("No block ranges stored, have you updated this client?");
        }
        var _a = this.getBundleState(latestBlockRange, tokenSymbol, chainId), flows = _a.flows, ubaConfig = _a.ubaConfig;
        return {
            flows: flows,
            ubaConfig: ubaConfig,
        };
    };
    UBAClientWithRefresh.prototype.getBundleState = function (blockRange, tokenSymbol, chainId) {
        var key = this.getKeyForBundle(blockRange, tokenSymbol, chainId);
        var bundleState = this.ubaBundleStates[key];
        if (!(0, utils_1.isDefined)(bundleState)) {
            throw new Error("Bundle state for chain ".concat(chainId, ", token ").concat(tokenSymbol, " and block range ").concat(blockRange, " should exist"));
        }
        return bundleState;
    };
    UBAClientWithRefresh.prototype.computeLpFee = function (baselineFee) {
        return baselineFee;
    };
    UBAClientWithRefresh.prototype.validateFlowsInBundle = function (blockRanges, tokenSymbol) {
        var _a;
        return tslib_1.__awaiter(this, void 0, void 0, function () {
            var l1TokenAddress, blockRangesForChain, flowsInBundle, _b, _i, flowsInBundle_1, flow, flowChain, validatedFlow;
            var _this = this;
            return tslib_1.__generator(this, function (_c) {
                switch (_c.label) {
                    case 0:
                        if (!this.ubaBundleBlockRanges.some(function (bundleBlockRanges) {
                            return JSON.stringify(blockRanges) === JSON.stringify(bundleBlockRanges);
                        })) {
                            throw new Error("Invalid block ranges not found in this.ubaBundleBlockRanges state");
                        }
                        l1TokenAddress = (_a = this.hubPoolClient.getL1Tokens().find(function (token) { return token.symbol === tokenSymbol; })) === null || _a === void 0 ? void 0 : _a.address;
                        if (!l1TokenAddress)
                            throw new Error("No L1 token address found for token symbol ".concat(tokenSymbol));
                        if ((0, utils_1.blockRangesAreInvalidForSpokeClients)(this.spokePoolClients, blockRanges, this.chainIdIndices)) {
                            throw new Error("Spoke pool clients do not have the block ranges necessary to look up data for bundle ranges: ".concat(JSON.stringify(blockRanges)));
                        }
                        blockRangesForChain = Object.fromEntries(this.enabledChainIds.map(function (chainId) {
                            return [chainId, (0, utils_1.getBlockRangeForChain)(blockRanges, chainId, _this.chainIdIndices)];
                        }));
                        _b = UBAClientUtilities_1.sortFlowsAscending;
                        return [4, (0, utils_1.mapAsync)(this.enabledChainIds, function (chainId) { return tslib_1.__awaiter(_this, void 0, void 0, function () {
                                var _a, startBlock, endBlock, flows;
                                return tslib_1.__generator(this, function (_b) {
                                    switch (_b.label) {
                                        case 0:
                                            _a = blockRangesForChain[chainId], startBlock = _a[0], endBlock = _a[1];
                                            if (startBlock === endBlock) {
                                                return [2, []];
                                            }
                                            return [4, (0, UBAClientUtilities_1.getUBAFlows)(tokenSymbol, chainId, this.spokePoolClients, this.hubPoolClient, startBlock, endBlock)];
                                        case 1:
                                            flows = _b.sent();
                                            return [2, flows];
                                    }
                                });
                            }); })];
                    case 1:
                        flowsInBundle = _b.apply(void 0, [(_c.sent()).flat()]);
                        _i = 0, flowsInBundle_1 = flowsInBundle;
                        _c.label = 2;
                    case 2:
                        if (!(_i < flowsInBundle_1.length)) return [3, 5];
                        flow = flowsInBundle_1[_i];
                        flowChain = (0, UBAClientUtilities_1.getFlowChain)(flow);
                        return [4, this.validateFlow(flow)];
                    case 3:
                        validatedFlow = _c.sent();
                        if ((0, utils_1.isDefined)(validatedFlow)) {
                            this.appendValidatedFlowsToClassState(flowChain, tokenSymbol, [validatedFlow], blockRanges);
                        }
                        else {
                        }
                        _c.label = 4;
                    case 4:
                        _i++;
                        return [3, 2];
                    case 5: return [2, Object.fromEntries(this.enabledChainIds.map(function (chainId) {
                            var bundleStateForChain = _this.getBundleState(blockRanges, tokenSymbol, chainId);
                            return [chainId, bundleStateForChain.flows];
                        }))];
                }
            });
        });
    };
    UBAClientWithRefresh.prototype.validateFlow = function (flow) {
        var _a, _b, _c, _d, _e, _f, _g;
        return tslib_1.__awaiter(this, void 0, void 0, function () {
            var flowChain, tokenSymbol, l1TokenAddress, blockRangesContainingFlow, bundleSharedState, openingBalanceForChain, ubaConfigForChain, precedingValidatedFlows, lastFlow, latestRunningBalance, latestIncentiveBalance, latestNetRunningBalanceAdjustment, potentialBalancingFee, _h, runningBalance, incentiveBalance, netRunningBalanceAdjustment, lpFeePct, lpFee, newModifiedFlow, matchedDeposit, bundleForMatchedDeposit, depositBundleState, ubaConfigForDeposit, lpFeePct_1, matchedDepositFlow, matchedDepositBlockRangeInFlowBundle, matchedDepositBundleFlows, validatedFlowsOnOriginChain, matchedDepositBundleFlows, expectedRealizedLpFeeForMatchedDeposit, expectedRealizedLpFeePctForMatchedDeposit;
            return tslib_1.__generator(this, function (_j) {
                switch (_j.label) {
                    case 0:
                        if ((0, interfaces_1.isUbaInflow)(flow)) {
                            flowChain = flow.originChainId;
                            tokenSymbol = (_a = this.hubPoolClient.getL1TokenInfoForL2Token(flow.originToken, flowChain)) === null || _a === void 0 ? void 0 : _a.symbol;
                        }
                        else {
                            if ((0, interfaces_1.outflowIsFill)(flow)) {
                                flowChain = flow.destinationChainId;
                                tokenSymbol = (_b = this.hubPoolClient.getL1TokenInfoForL2Token(flow.destinationToken, flowChain)) === null || _b === void 0 ? void 0 : _b.symbol;
                            }
                            else {
                                flowChain = flow.repaymentChainId;
                                tokenSymbol = (_c = this.hubPoolClient.getL1TokenInfoForL2Token(flow.refundToken, flowChain)) === null || _c === void 0 ? void 0 : _c.symbol;
                            }
                        }
                        if (!(0, utils_1.isDefined)(tokenSymbol))
                            throw new Error("No token symbol found");
                        l1TokenAddress = (_d = this.hubPoolClient.getL1Tokens().find(function (token) { return token.symbol === tokenSymbol; })) === null || _d === void 0 ? void 0 : _d.address;
                        if (!(0, utils_1.isDefined)(l1TokenAddress))
                            throw new Error("No L1 token address found for token symbol");
                        blockRangesContainingFlow = this.getUbaBundleBlockRangeContainingFlow(flow.blockNumber, flowChain);
                        bundleSharedState = this.getBundleState(blockRangesContainingFlow, tokenSymbol, flowChain);
                        openingBalanceForChain = bundleSharedState.openingBalances;
                        ubaConfigForChain = bundleSharedState.ubaConfig;
                        precedingValidatedFlows = bundleSharedState.flows;
                        lastFlow = precedingValidatedFlows[precedingValidatedFlows.length - 1];
                        latestRunningBalance = (_e = lastFlow === null || lastFlow === void 0 ? void 0 : lastFlow.runningBalance) !== null && _e !== void 0 ? _e : openingBalanceForChain.runningBalance;
                        latestIncentiveBalance = (_f = lastFlow === null || lastFlow === void 0 ? void 0 : lastFlow.incentiveBalance) !== null && _f !== void 0 ? _f : openingBalanceForChain.incentiveBalance;
                        latestNetRunningBalanceAdjustment = (_g = lastFlow === null || lastFlow === void 0 ? void 0 : lastFlow.netRunningBalanceAdjustment) !== null && _g !== void 0 ? _g : ethers_1.ethers.constants.Zero;
                        if ((0, interfaces_1.isUbaInflow)(flow)) {
                            (potentialBalancingFee = (0, UBAFeeSpokeCalculatorAnalog_1.getDepositFee)(flow.amount, latestRunningBalance, latestIncentiveBalance, flowChain, ubaConfigForChain).balancingFee);
                        }
                        else {
                            (potentialBalancingFee = (0, UBAFeeSpokeCalculatorAnalog_1.getRefundFee)(flow.amount, latestRunningBalance, latestIncentiveBalance, flowChain, ubaConfigForChain).balancingFee);
                        }
                        _h = UBAFeeCalculator_1.analog.calculateHistoricalRunningBalance([
                            tslib_1.__assign(tslib_1.__assign({}, flow), { incentiveFee: potentialBalancingFee }),
                        ], latestRunningBalance, latestIncentiveBalance, latestNetRunningBalanceAdjustment, flowChain, tokenSymbol, ubaConfigForChain), runningBalance = _h.runningBalance, incentiveBalance = _h.incentiveBalance, netRunningBalanceAdjustment = _h.netRunningBalanceAdjustment;
                        lpFeePct = this.computeLpFee(ubaConfigForChain.getBaselineFee(flow.destinationChainId, flow.originChainId));
                        lpFee = lpFeePct.mul(flow.amount).div(utils_1.fixedPointAdjustment);
                        newModifiedFlow = {
                            flow: flow,
                            balancingFee: potentialBalancingFee,
                            lpFee: lpFee,
                            runningBalance: runningBalance,
                            incentiveBalance: incentiveBalance,
                            netRunningBalanceAdjustment: netRunningBalanceAdjustment,
                        };
                        if (!(0, interfaces_1.isUbaInflow)(flow)) return [3, 1];
                        return [2, newModifiedFlow];
                    case 1:
                        if (!(0, utils_1.isDefined)(flow.realizedLpFeePct)) {
                            throw new Error("Outflow has undefined realizedLpFeePct");
                        }
                        if ((0, utils_1.isDefined)(flow.matchedDeposit.realizedLpFeePct)) {
                            if (flow.matchedDeposit.realizedLpFeePct.eq(flow.realizedLpFeePct)) {
                                return [2, {
                                        flow: flow,
                                        runningBalance: latestRunningBalance,
                                        incentiveBalance: latestIncentiveBalance,
                                        netRunningBalanceAdjustment: latestNetRunningBalanceAdjustment,
                                        balancingFee: ethers_1.ethers.constants.Zero,
                                        lpFee: flow.realizedLpFeePct.mul(flow.amount).div(utils_1.fixedPointAdjustment),
                                    }];
                            }
                            else {
                                this.logger.debug({
                                    at: "UBAClientWithRefresh#validateFlow",
                                    message: "Flow matched a pre-UBA deposit, but it was invalid because it set the wrong LP fee",
                                    flow: flow,
                                });
                                return [2, undefined];
                            }
                        }
                        matchedDeposit = flow.matchedDeposit;
                        bundleForMatchedDeposit = this.getUbaBundleBlockRangeContainingFlow(matchedDeposit.blockNumber, matchedDeposit.originChainId);
                        depositBundleState = this.getBundleState(bundleForMatchedDeposit, tokenSymbol, matchedDeposit.originChainId);
                        ubaConfigForDeposit = depositBundleState.ubaConfig;
                        if (ubaConfigForDeposit.isBalancingFeeCurveFlatAtZero(flow.matchedDeposit.originChainId)) {
                            lpFeePct_1 = newModifiedFlow.lpFee.mul(utils_1.fixedPointAdjustment).div(flow.amount);
                            if (!lpFeePct_1.eq(flow.realizedLpFeePct)) {
                                return [2, undefined];
                            }
                            else {
                                return [2, newModifiedFlow];
                            }
                        }
                        if (!(flow.blockTimestamp < flow.matchedDeposit.blockTimestamp)) return [3, 2];
                        if (this.latestBlockTimestamps[flow.matchedDeposit.originChainId] < flow.blockTimestamp) {
                            this.logger.error({
                                at: "UBAClientWithRefresh#validateFlow",
                                message: "We cannot invalidate a fill if the latest block.timestamp on the origin chain is <= the fill's timestamp",
                                latestOriginChainBlockTimestamp: this.latestBlockTimestamps[flow.matchedDeposit.originChainId],
                                flow: flow,
                            });
                            throw new Error("We cannot invalidate a fill if the latest block.timestamp ".concat(this.latestBlockTimestamps[flow.matchedDeposit.originChainId], " on the deposit.origin chain is not > than the fill's timestamp ").concat(flow.blockTimestamp, ". This is because its still possible to send a deposit on the origin chain that would validate this fill."));
                        }
                        return [3, 10];
                    case 2:
                        matchedDepositFlow = void 0;
                        matchedDepositBlockRangeInFlowBundle = (0, utils_1.getBlockRangeForChain)(blockRangesContainingFlow, matchedDeposit.originChainId, this.chainIdIndices);
                        if (!(matchedDeposit.blockNumber > matchedDepositBlockRangeInFlowBundle[1])) return [3, 6];
                        matchedDepositFlow = this.getMatchingValidatedFlow(matchedDeposit.blockNumber, matchedDeposit.originChainId, matchedDeposit, tokenSymbol);
                        if (!(0, utils_1.isDefined)(matchedDepositFlow)) return [3, 3];
                        return [3, 5];
                    case 3: return [4, this.validateFlowsInBundle(bundleForMatchedDeposit, tokenSymbol)];
                    case 4:
                        matchedDepositBundleFlows = _j.sent();
                        matchedDepositFlow = (0, UBAClientUtilities_1.getMatchingFlow)(matchedDepositBundleFlows[matchedDeposit.originChainId], matchedDeposit);
                        _j.label = 5;
                    case 5: return [3, 9];
                    case 6:
                        if (!(matchedDeposit.blockNumber >= matchedDepositBlockRangeInFlowBundle[0] &&
                            matchedDeposit.blockNumber <= matchedDepositBlockRangeInFlowBundle[1])) return [3, 7];
                        validatedFlowsOnOriginChain = depositBundleState.flows;
                        matchedDepositFlow = (0, UBAClientUtilities_1.getMatchingFlow)(validatedFlowsOnOriginChain, matchedDeposit);
                        if (!(0, utils_1.isDefined)(matchedDepositFlow)) {
                            this.logger.error({
                                at: "UBAClientWithRefresh#validateFlow",
                                message: "Deposit should be in same bundle as flow and already have been validated but cannot be found",
                                validatedFlowsOnOriginChain: validatedFlowsOnOriginChain,
                                flow: flow,
                            });
                            throw new Error("Could not find matched deposit in same bundle as fill");
                        }
                        return [3, 9];
                    case 7:
                        matchedDepositFlow = this.getMatchingValidatedFlow(matchedDeposit.blockNumber, matchedDeposit.originChainId, matchedDeposit, tokenSymbol);
                        if (!!(0, utils_1.isDefined)(matchedDepositFlow)) return [3, 9];
                        return [4, this.validateFlowsInBundle(bundleForMatchedDeposit, tokenSymbol)];
                    case 8:
                        matchedDepositBundleFlows = _j.sent();
                        matchedDepositFlow = (0, UBAClientUtilities_1.getMatchingFlow)(matchedDepositBundleFlows[matchedDeposit.originChainId], matchedDeposit);
                        _j.label = 9;
                    case 9:
                        if (!(0, utils_1.isDefined)(matchedDepositFlow)) {
                            this.logger.error({
                                at: "UBAClientWithRefresh#validateFlow",
                                message: "No matching flow found in client state for matched deposit",
                                flow: flow,
                            });
                            throw new Error("Could not validate or invalidate matched deposit");
                        }
                        expectedRealizedLpFeeForMatchedDeposit = matchedDepositFlow.lpFee.add(matchedDepositFlow.balancingFee);
                        expectedRealizedLpFeePctForMatchedDeposit = expectedRealizedLpFeeForMatchedDeposit
                            .mul(utils_1.fixedPointAdjustment)
                            .div(matchedDepositFlow.flow.amount);
                        if (expectedRealizedLpFeePctForMatchedDeposit.lte(flow.realizedLpFeePct.mul((0, utils_1.toBNWei)("1.01")).div(utils_1.fixedPointAdjustment)) ||
                            expectedRealizedLpFeePctForMatchedDeposit.gte(flow.realizedLpFeePct.mul((0, utils_1.toBNWei)("0.99")).div(utils_1.fixedPointAdjustment))) {
                            return [2, newModifiedFlow];
                        }
                        else {
                        }
                        _j.label = 10;
                    case 10: return [2, undefined];
                }
            });
        });
    };
    UBAClientWithRefresh.prototype.getMatchingValidatedFlow = function (flowBlock, flowChain, flow, tokenSymbol) {
        var bundleContainingFlow = this.getUbaBundleBlockRangeContainingFlow(flowBlock, flowChain);
        var validatedFlowsInBundle = this.getBundleState(bundleContainingFlow, tokenSymbol, flowChain).flows;
        var matchingFlow = (0, UBAClientUtilities_1.getMatchingFlow)(validatedFlowsInBundle, flow);
        return matchingFlow;
    };
    UBAClientWithRefresh.prototype.getMostRecentBundleBlockRangesPerChain = function (bundleCount) {
        var _this = this;
        return this.chainIdIndices.reduce(function (acc, chainId) {
            var _a;
            var _blockRangesForChain = (0, UBAClientUtilities_1.getMostRecentBundleBlockRanges)(chainId, bundleCount, _this.hubPoolClient, _this.spokePoolClients).map(function (_a) {
                var start = _a.start, end = _a.end;
                return [start, end];
            });
            var ubaActivationBundleStartBlockForChain = (0, utils_1.getBlockForChain)((0, UBAClientUtilities_1.getUbaActivationBundleStartBlocks)(_this.hubPoolClient), chainId, _this.chainIdIndices);
            if (_blockRangesForChain.length === 0) {
                throw new Error("Should never return 0 length block ranges for chain ".concat(chainId));
            }
            if (_blockRangesForChain[0][0] !== ubaActivationBundleStartBlockForChain ||
                ((0, utils_1.isDefined)(_this.spokePoolClients[chainId]) &&
                    _blockRangesForChain.at(-1)[1] !== _this.spokePoolClients[chainId].latestBlockSearched)) {
                _this.logger.error({
                    at: "UBAClientWithRefresh#getMostRecentBundleBlockRangesPerChain",
                    message: "Block ranges for chain ".concat(chainId, " do not cover from UBA activation bundle start block to latest spoke pool client block searched"),
                    startBlockForChain: _blockRangesForChain[0][0],
                    ubaActivationBundleStartBlockForChain: ubaActivationBundleStartBlockForChain,
                    endBlockForChain: (_a = _blockRangesForChain.at(-1)) === null || _a === void 0 ? void 0 : _a[1],
                    latestSpokePoolClientBlockSearched: _this.spokePoolClients[chainId].latestBlockSearched,
                });
                throw new Error("Block ranges for chain ".concat(chainId, " do not cover from UBA activation bundle start block to latest spoke pool client block searched"));
            }
            acc[chainId] = _blockRangesForChain;
            return acc;
        }, {});
    };
    UBAClientWithRefresh.prototype.getUbaBundleBlockRangeContainingFlow = function (flowBlock, flowChain) {
        var _this = this;
        var blockRangesContainingFlow = this.ubaBundleBlockRanges.find(function (blockRanges) {
            var blockRangeForChain = (0, utils_1.getBlockRangeForChain)(blockRanges, flowChain, _this.chainIdIndices);
            return blockRangeForChain[0] <= flowBlock && flowBlock <= blockRangeForChain[1];
        });
        if (!blockRangesContainingFlow) {
            throw new Error("Could not find bundle block range containing flow at block ".concat(flowBlock, " on chain ").concat(flowChain));
        }
        return blockRangesContainingFlow;
    };
    UBAClientWithRefresh.prototype.appendValidatedFlowsToClassState = function (chainId, tokenSymbol, flows, bundleBlockRanges) {
        var _a;
        var bundleState = this.getBundleState(bundleBlockRanges, tokenSymbol, chainId);
        (_a = bundleState.flows).push.apply(_a, flows);
    };
    UBAClientWithRefresh.prototype.update = function () {
        return tslib_1.__awaiter(this, void 0, void 0, function () {
            var latestHubPoolBlock, tokens, blockRangesByChain, bundleBlockRangesCount, bundleBlockRanges, _loop_1, this_1, i, _loop_2, this_2, i, latestTimestampsPerChain, _loop_3, this_3, i, newUbaClientState, _loop_4, this_4, i, _loop_5, this_5, i, _loop_6, i;
            var _this = this;
            return tslib_1.__generator(this, function (_a) {
                switch (_a.label) {
                    case 0:
                        latestHubPoolBlock = this.hubPoolClient.latestBlockSearched;
                        this.logger.debug({
                            at: "UBAClientWithRefresh",
                            message: "❤️‍🔥😭  Updating UBA Client",
                            enabledChains: this.enabledChainIds,
                        });
                        tokens = this.tokens;
                        this.logger.debug({
                            at: "UBAClientWithRefresh#getMostRecentBundleBlockRangesPerChain",
                            message: "Loaded UBA bundle start blocks",
                            ubaActivationBundleStartBlockForChain: (0, UBAClientUtilities_1.getUbaActivationBundleStartBlocks)(this.hubPoolClient),
                        });
                        blockRangesByChain = this.getMostRecentBundleBlockRangesPerChain(100);
                        bundleBlockRangesCount = blockRangesByChain[this.chainIdIndices[0]].length;
                        bundleBlockRanges = [];
                        _loop_1 = function (i) {
                            bundleBlockRanges.push(this_1.chainIdIndices
                                .map(function (chainId) {
                                var blockRangeCountForChain = blockRangesByChain[chainId].length;
                                if (blockRangeCountForChain < bundleBlockRangesCount) {
                                    var firstBlockRange = blockRangesByChain[chainId][0];
                                    if (i < bundleBlockRangesCount - blockRangeCountForChain) {
                                        return [firstBlockRange[0], firstBlockRange[0]];
                                    }
                                    else {
                                        return firstBlockRange;
                                    }
                                }
                                return blockRangesByChain[chainId][i];
                            })
                                .filter(utils_1.isDefined));
                        };
                        this_1 = this;
                        for (i = 0; i < bundleBlockRangesCount; i++) {
                            _loop_1(i);
                        }
                        if (bundleBlockRanges.some(function (blockRange) { return blockRange.length !== _this.chainIdIndices.length; })) {
                            throw new Error("Block ranges are not equal length for all bundles");
                        }
                        this.ubaBundleBlockRanges = bundleBlockRanges;
                        this.logger.debug({
                            at: "UBAClientWithRefresh#update",
                            message: "UBA bundle block ranges we're storing in class memory",
                            bundleBlockRanges: this.ubaBundleBlockRanges,
                        });
                        _loop_2 = function (i) {
                            var bundleBlockRange = this_2.ubaBundleBlockRanges[i];
                            var startBlocks = bundleBlockRange.map(function (_a) {
                                var start = _a[0];
                                return start;
                            });
                            var mainnetStartBlock = (0, utils_1.getBlockForChain)(startBlocks, this_2.hubPoolClient.chainId, this_2.chainIdIndices);
                            tokens.forEach(function (token) {
                                var _a;
                                var l1TokenAddress = (_a = _this.hubPoolClient.getL1Tokens().find(function (_token) { return _token.symbol === token; })) === null || _a === void 0 ? void 0 : _a.address;
                                if (!(0, utils_1.isDefined)(l1TokenAddress))
                                    throw new Error("No L1 token address found for token symbol ".concat(token));
                                _this.enabledChainIds.forEach(function (chainId) {
                                    var bundleStateKey = _this.getKeyForBundle(bundleBlockRange, token, chainId);
                                    var startBlock = (0, utils_1.getBlockForChain)(startBlocks, chainId, _this.chainIdIndices);
                                    var openingBalances = (0, UBAClientUtilities_1.getOpeningRunningBalanceForEvent)(_this.hubPoolClient, startBlock, chainId, l1TokenAddress, latestHubPoolBlock);
                                    var ubaConfig = (0, UBAClientUtilities_1.getUBAFeeConfig)(_this.hubPoolClient, chainId, token, mainnetStartBlock);
                                    _this.ubaBundleStates[bundleStateKey] = {
                                        openingBalances: openingBalances,
                                        ubaConfig: ubaConfig,
                                        flows: [],
                                        loadedFromCache: false,
                                    };
                                });
                            });
                        };
                        this_2 = this;
                        for (i = 0; i < this.ubaBundleBlockRanges.length; i++) {
                            _loop_2(i);
                        }
                        latestTimestampsPerChain = Object.fromEntries(this.enabledChainIds
                            .map(function (chainId) {
                            if (!(0, utils_1.isDefined)(_this.spokePoolClients[chainId]))
                                return undefined;
                            return [chainId, _this.spokePoolClients[chainId].getCurrentTime()];
                        })
                            .filter(utils_1.isDefined));
                        this.latestBlockTimestamps = latestTimestampsPerChain;
                        this.logger.debug({
                            at: "UBAClientWithRefresh#update",
                            message: "Latest block timestamps per chain",
                            latestBlockTimestamps: this.latestBlockTimestamps,
                        });
                        if (!(0, utils_1.isDefined)(this.cachingMechanism)) return [3, 4];
                        _loop_3 = function (i) {
                            var mostRecentBundleBlockRanges, mainnetBundleStartBlock;
                            return tslib_1.__generator(this, function (_b) {
                                switch (_b.label) {
                                    case 0:
                                        mostRecentBundleBlockRanges = this_3.ubaBundleBlockRanges[i];
                                        mainnetBundleStartBlock = (0, utils_1.getBlockRangeForChain)(mostRecentBundleBlockRanges, this_3.hubPoolClient.chainId, this_3.chainIdIndices)[0];
                                        if ((0, utils_1.isDefined)(this_3.clientConfig.latestMainnetBundleStartBlockToLoadFromCache) &&
                                            this_3.clientConfig.latestMainnetBundleStartBlockToLoadFromCache > mainnetBundleStartBlock) {
                                            this_3.logger.debug({
                                                at: "UBAClientWithRefresh#update",
                                                message: "Skipping loading bundle data from cache for bundle with mainnet start block ".concat(mainnetBundleStartBlock, " because it's older than the oldest bundle we want to load from the cache"),
                                                bundleBlockRanges: mostRecentBundleBlockRanges,
                                                oldestMainnetStartBlockToLoadFromCache: this_3.clientConfig.latestMainnetBundleStartBlockToLoadFromCache,
                                            });
                                            return [2, "continue"];
                                        }
                                        return [4, (0, utils_1.forEachAsync)(tokens, function (token) { return tslib_1.__awaiter(_this, void 0, void 0, function () {
                                                var _this = this;
                                                return tslib_1.__generator(this, function (_a) {
                                                    switch (_a.label) {
                                                        case 0: return [4, (0, utils_1.forEachAsync)(this.enabledChainIds, function (chainId) { return tslib_1.__awaiter(_this, void 0, void 0, function () {
                                                                var redisKeyForBundle, modifiedFlowsInBundle;
                                                                var _a;
                                                                return tslib_1.__generator(this, function (_b) {
                                                                    switch (_b.label) {
                                                                        case 0:
                                                                            redisKeyForBundle = this.getKeyForBundle(mostRecentBundleBlockRanges, token, chainId);
                                                                            return [4, ((_a = this.cachingMechanism) === null || _a === void 0 ? void 0 : _a.get(redisKeyForBundle))];
                                                                        case 1:
                                                                            modifiedFlowsInBundle = _b.sent();
                                                                            if ((0, utils_1.isDefined)(modifiedFlowsInBundle)) {
                                                                                this.logger.debug({
                                                                                    at: "UBAClientWithRefresh#update",
                                                                                    message: "\uD83D\uDCBF Loaded bundle state from cache using key ".concat(redisKeyForBundle),
                                                                                    bundleBlockRanges: mostRecentBundleBlockRanges,
                                                                                    flowsLoaded: modifiedFlowsInBundle.length,
                                                                                });
                                                                                this.appendValidatedFlowsToClassState(chainId, token, modifiedFlowsInBundle, mostRecentBundleBlockRanges);
                                                                                this.getBundleState(mostRecentBundleBlockRanges, token, chainId).loadedFromCache = true;
                                                                            }
                                                                            else {
                                                                                this.logger.debug({
                                                                                    at: "UBAClientWithRefresh#update",
                                                                                    message: "No entry for key ".concat(redisKeyForBundle, " in redis"),
                                                                                    bundleBlockRanges: mostRecentBundleBlockRanges,
                                                                                });
                                                                            }
                                                                            return [2];
                                                                    }
                                                                });
                                                            }); })];
                                                        case 1:
                                                            _a.sent();
                                                            return [2];
                                                    }
                                                });
                                            }); })];
                                    case 1:
                                        _b.sent();
                                        return [2];
                                }
                            });
                        };
                        this_3 = this;
                        i = this.ubaBundleBlockRanges.length - 1;
                        _a.label = 1;
                    case 1:
                        if (!(i >= 0)) return [3, 4];
                        return [5, _loop_3(i)];
                    case 2:
                        _a.sent();
                        _a.label = 3;
                    case 3:
                        i--;
                        return [3, 1];
                    case 4:
                        newUbaClientState = {};
                        _loop_4 = function (i) {
                            var mostRecentBundleBlockRanges;
                            return tslib_1.__generator(this, function (_c) {
                                switch (_c.label) {
                                    case 0:
                                        mostRecentBundleBlockRanges = this_4.ubaBundleBlockRanges[i];
                                        return [4, (0, utils_1.forEachAsync)(tokens, function (token) { return tslib_1.__awaiter(_this, void 0, void 0, function () {
                                                var modifiedFlowsInBundle;
                                                var _this = this;
                                                return tslib_1.__generator(this, function (_a) {
                                                    switch (_a.label) {
                                                        case 0:
                                                            if (!this.enabledChainIds.every(function (chainId) {
                                                                return _this.getBundleState(mostRecentBundleBlockRanges, token, chainId).loadedFromCache;
                                                            })) return [3, 1];
                                                            modifiedFlowsInBundle = Object.fromEntries(this.enabledChainIds.map(function (chainId) {
                                                                var cachedFlows = _this.getBundleState(mostRecentBundleBlockRanges, token, chainId).flows;
                                                                return [chainId, cachedFlows];
                                                            }));
                                                            return [3, 3];
                                                        case 1: return [4, this.validateFlowsInBundle(mostRecentBundleBlockRanges, token)];
                                                        case 2:
                                                            modifiedFlowsInBundle = _a.sent();
                                                            _a.label = 3;
                                                        case 3: return [4, (0, utils_1.forEachAsync)(this.enabledChainIds, function (chainId) { return tslib_1.__awaiter(_this, void 0, void 0, function () {
                                                                var bundleState, redisKeyForBundle;
                                                                return tslib_1.__generator(this, function (_a) {
                                                                    switch (_a.label) {
                                                                        case 0:
                                                                            if (!(0, utils_1.isDefined)(newUbaClientState[chainId]))
                                                                                newUbaClientState[chainId] = {};
                                                                            if (!(0, utils_1.isDefined)(newUbaClientState[chainId][token]))
                                                                                newUbaClientState[chainId][token] = [];
                                                                            bundleState = this.getBundleState(mostRecentBundleBlockRanges, token, chainId);
                                                                            newUbaClientState[chainId][token].push({
                                                                                bundleBlockRanges: mostRecentBundleBlockRanges,
                                                                                flows: bundleState.flows,
                                                                                ubaConfig: bundleState.ubaConfig,
                                                                                openingBalances: bundleState.openingBalances,
                                                                            });
                                                                            if (this.ubaBundleBlockRanges.length <= 30)
                                                                                return [2];
                                                                            if (i === this.ubaBundleBlockRanges.length - 1)
                                                                                return [2];
                                                                            redisKeyForBundle = this.getKeyForBundle(mostRecentBundleBlockRanges, token, chainId);
                                                                            if (!(0, utils_1.isDefined)(this.cachingMechanism)) return [3, 2];
                                                                            return [4, this.cachingMechanism.set(redisKeyForBundle, modifiedFlowsInBundle[chainId])];
                                                                        case 1:
                                                                            _a.sent();
                                                                            _a.label = 2;
                                                                        case 2: return [2];
                                                                    }
                                                                });
                                                            }); })];
                                                        case 4:
                                                            _a.sent();
                                                            return [2];
                                                    }
                                                });
                                            }); })];
                                    case 1:
                                        _c.sent();
                                        return [2];
                                }
                            });
                        };
                        this_4 = this;
                        i = 0;
                        _a.label = 5;
                    case 5:
                        if (!(i < this.ubaBundleBlockRanges.length)) return [3, 8];
                        return [5, _loop_4(i)];
                    case 6:
                        _a.sent();
                        _a.label = 7;
                    case 7:
                        i++;
                        return [3, 5];
                    case 8:
                        _loop_5 = function (i) {
                            var mostRecentBundleBlockRanges = this_5.ubaBundleBlockRanges[i];
                            this_5.enabledChainIds.forEach(function (chainId) {
                                tokens.forEach(function (token) {
                                    var modifiedFlowsInBundle = _this.getBundleState(mostRecentBundleBlockRanges, token, chainId).flows;
                                    modifiedFlowsInBundle.forEach(function (_a) {
                                        var flow = _a.flow;
                                        if ((0, interfaces_1.isUbaOutflow)(flow) && !(0, utils_1.isDefined)(flow.matchedDeposit.realizedLpFeePct)) {
                                            _this.spokePoolClients[flow.matchedDeposit.originChainId].updateDepositRealizedLpFeePct(flow.matchedDeposit, flow.realizedLpFeePct);
                                        }
                                    });
                                });
                            });
                        };
                        this_5 = this;
                        for (i = 0; i < this.ubaBundleBlockRanges.length; i++) {
                            _loop_5(i);
                        }
                        this.ubaClientState = newUbaClientState;
                        this.isUpdated = true;
                        _loop_6 = function (i) {
                            tokens.forEach(function (token) {
                                var bundleBlockRange = _this.ubaBundleBlockRanges[i];
                                var breakdownPerChain = _this.enabledChainIds
                                    .map(function (chainId) {
                                    var bundleState = _this.getBundleState(bundleBlockRange, token, chainId);
                                    if ((0, utils_1.isDefined)(bundleState)) {
                                        var flows = bundleState.flows;
                                        var lastFlow = flows.at(-1);
                                        if (!(0, utils_1.isDefined)(lastFlow))
                                            return undefined;
                                        var fills = flows.filter(function (_a) {
                                            var flow = _a.flow;
                                            return (0, interfaces_1.isUbaOutflow)(flow) && (0, interfaces_1.outflowIsFill)(flow) && flow.matchedDeposit.realizedLpFeePct === undefined;
                                        });
                                        var deposits = flows.filter(function (_a) {
                                            var flow = _a.flow;
                                            return (0, interfaces_1.isUbaInflow)(flow);
                                        });
                                        var refunds = flows.filter(function (_a) {
                                            var flow = _a.flow;
                                            return (0, interfaces_1.isUbaOutflow)(flow) && (0, interfaces_1.outflowIsRefund)(flow) && flow.matchedDeposit.realizedLpFeePct === undefined;
                                        });
                                        var inflows = deposits.reduce(function (sum, _a) {
                                            var flow = _a.flow;
                                            sum = sum.add(flow.amount);
                                            return sum;
                                        }, ethers_1.ethers.constants.Zero);
                                        var fillOutflows = fills.reduce(function (sum, _a) {
                                            var flow = _a.flow;
                                            sum = sum.add(flow.amount);
                                            return sum;
                                        }, ethers_1.ethers.constants.Zero);
                                        var refundOutflows = flows.reduce(function (sum, _a) {
                                            var balancingFee = _a.balancingFee;
                                            return sum.add(balancingFee);
                                        }, ethers_1.ethers.constants.Zero);
                                        var balancingFees = flows.reduce(function (sum, _a) {
                                            var balancingFee = _a.balancingFee;
                                            return sum.add(balancingFee);
                                        }, ethers_1.ethers.constants.Zero);
                                        var readableFlows = {
                                            fills: fills.length,
                                            fillRunningBalanceRemoved: fillOutflows.mul(-1).toString(),
                                            deposits: deposits.length,
                                            depositRunningBalanceAdded: inflows.toString(),
                                            refunds: refunds.length,
                                            refundRunningBalanceRemoved: refundOutflows.mul(-1).toString(),
                                            balancingFeesAdded: balancingFees.toString(),
                                            lpFeesAdded: flows
                                                .reduce(function (sum, _a) {
                                                var lpFee = _a.lpFee;
                                                return sum.add(lpFee);
                                            }, ethers_1.ethers.constants.Zero)
                                                .toString(),
                                            netRunningBalanceAdjustment: lastFlow.netRunningBalanceAdjustment.toString(),
                                            openingRunningBalance: bundleState.openingBalances.runningBalance.toString(),
                                            closingRunningBalance: lastFlow.runningBalance.toString(),
                                            openingIncentiveBalance: bundleState.openingBalances.incentiveBalance.toString(),
                                            closingIncentiveBalance: lastFlow.incentiveBalance.toString(),
                                        };
                                        var expectedClosingBalance = bundleState.openingBalances.runningBalance
                                            .add(inflows)
                                            .sub(refundOutflows)
                                            .sub(fillOutflows)
                                            .sub(balancingFees)
                                            .add(lastFlow.netRunningBalanceAdjustment);
                                        if (!expectedClosingBalance.eq(lastFlow.runningBalance)) {
                                            _this.logger.error({
                                                at: "UBAClientWithRefresh#update",
                                                message: "Expected closing balance ".concat(expectedClosingBalance.toString(), " to equal actual closing balance ").concat(lastFlow.runningBalance.toString()),
                                                bundleBlockRange: bundleBlockRange,
                                                chainId: chainId,
                                                token: token,
                                                openingRunningBalance: bundleState.openingBalances.runningBalance.toString(),
                                                inflows: inflows.toString(),
                                                fillOutflows: fillOutflows.mul(-1).toString(),
                                                refundOutflows: refundOutflows.mul(-1).toString(),
                                                balancingFees: balancingFees.mul(-1).toString(),
                                                netRunningBalanceAdjustment: lastFlow.netRunningBalanceAdjustment.toString(),
                                            });
                                        }
                                        return [chainId, readableFlows];
                                    }
                                    else
                                        return undefined;
                                })
                                    .filter(utils_1.isDefined);
                                var breakdown = Object.fromEntries(breakdownPerChain);
                                _this.logger.debug({
                                    at: "UBAClientWithRefresh#update",
                                    message: "Bundle state for ".concat(token),
                                    bundleBlockRange: bundleBlockRange,
                                    breakdown: breakdown,
                                });
                            });
                        };
                        for (i = 0; i < this.ubaBundleBlockRanges.length; i++) {
                            _loop_6(i);
                        }
                        return [2];
                }
            });
        });
    };
    return UBAClientWithRefresh;
}(BaseAbstractClient_1.BaseAbstractClient));
exports.UBAClientWithRefresh = UBAClientWithRefresh;
//# sourceMappingURL=UBAClientWithRefresh.js.map