import { __assign, __awaiter, __extends, __generator } from "tslib";
import assert from "assert";
import { getBundleKeyForBlockRanges, getFlowChain, getUBAFlows, getMatchingFlow, getMostRecentBundleBlockRanges, getOpeningRunningBalanceForEvent, getUBAFeeConfig, getUbaActivationBundleStartBlocks, sortFlowsAscending, } from "./UBAClientUtilities";
import { isUbaInflow, isUbaOutflow, outflowIsFill, outflowIsRefund, } from "../../interfaces";
import { blockRangesAreInvalidForSpokeClients, fixedPointAdjustment, forEachAsync, getBlockForChain, getBlockRangeForChain, isDefined, mapAsync, toBNWei, } from "../../utils";
import { analog } from "../../UBAFeeCalculator";
import { getDepositFee, getRefundFee } from "../../UBAFeeCalculator/UBAFeeSpokeCalculatorAnalog";
import { ethers } from "ethers";
import { BaseAbstractClient } from "../BaseAbstractClient";
/**
 * @notice This class reconstructs UBA bundle data for every bundle that has been created since the
 * UBA activation. Its performance can be signficantly sped up by integrating an external caching layer
 * like IPFS or Redis. It provides an interface that is designed to be called by actors who want to
 * to reconstruct a past bundle's state or validate a live bundle's state in order to evaluate a pending
 * root bundle, propose a new root bundle, or get the latest prices as a relayer or fee quoter.
 */
var UBAClientWithRefresh = /** @class */ (function (_super) {
    __extends(UBAClientWithRefresh, _super);
    /**
     * @notice Constructs a new UBAClientinstance.
     * @param clientConfig Configuration to customize the UBAClient instance.
     * @param tokens Tokens to load bundle state for.
     * @param hubPoolClient
     * @param spokePoolClients
     */
    function UBAClientWithRefresh(clientConfig, tokens, hubPoolClient, spokePoolClients, cachingClient) {
        var _this = _super.call(this, cachingClient) || this;
        _this.clientConfig = clientConfig;
        _this.tokens = tokens;
        _this.hubPoolClient = hubPoolClient;
        _this.spokePoolClients = spokePoolClients;
        // A public, convenient variable storing `ubaBundleStates` in a different format, with all bundle data
        // for a token and chain grouped together in a list.
        _this.ubaClientState = {};
        // Associates a single bundle state with a unique key. The key should be created via
        // `this.getKeyForBundle(blockRanges, tokenSymbol, chain)`.
        // Bundle states from older, already validated, bundles should ideally be seeded directly
        // from an external storage layer rather than reconstructed fresh from this class. These state should
        // be populated for all block ranges where the UBA is active after calling update().
        _this.ubaBundleStates = {};
        // All bundle ranges loaded by this client. Should contain all bundle ranges for all validated bundles
        // as of the UBA activation block that the hub pool client is aware of. This is fast to load
        // since it only depends on HubPoolClient event history. These should be stores in chronologically
        // ascending order.
        _this.ubaBundleBlockRanges = [];
        _this.latestBlockTimestamps = {};
        _this.getKeyForBundle = function (bundleBlockRanges, tokenSymbol, chainId) {
            // Add a fixed prefix to all keys so we can more easily flush all bundle state keys.
            return "UBA_BUNDLE_STATE_".concat(getBundleKeyForBlockRanges(bundleBlockRanges), "-").concat(tokenSymbol, "-").concat(chainId);
        };
        _this.logger = _this.hubPoolClient.logger;
        _this.enabledChainIds = _this.hubPoolClient.configStoreClient.getEnabledChains();
        assert(_this.enabledChainIds.length > 0, "No chainIds provided");
        _this.enabledChainIds.forEach(function (chainId) {
            assert(isDefined(spokePoolClients[chainId]), "No SpokePool provided for chainId ".concat(chainId));
        });
        _this.chainIdIndices = _this.hubPoolClient.configStoreClient.getChainIdIndicesForBlock();
        return _this;
    }
    // /////////////////
    //
    //
    // PUBLIC METHODS:
    //
    // /////////////////
    /**
     * @notice Intended to be called by Relayer to set `realizedLpFeePct` for a deposit, where
     * realizedLpFeePct for this deposit should be equal to the lpFee plus the depositBalancingFee.
     * @param deposit Should be the deposit that the caller wants to compute the realizedLpFeePct for.
     */
    UBAClientWithRefresh.prototype.computeFeesForDeposit = function (deposit) {
        var _a;
        this.assertUpdated();
        var tokenSymbol = (_a = this.hubPoolClient.getL1TokenInfoForL2Token(deposit.originToken, deposit.originChainId)) === null || _a === void 0 ? void 0 : _a.symbol;
        if (!tokenSymbol)
            throw new Error("No token symbol found");
        // Grab bundle state containing deposit using the bundle state's block ranges.
        var blockRangeContainingDeposit = this.getUbaBundleBlockRangeContainingFlow(deposit.blockNumber, deposit.originChainId);
        var specificBundleState = this.getBundleState(blockRangeContainingDeposit, tokenSymbol, deposit.originChainId);
        // Find matching flow in bundle state.
        var matchingFlow = getMatchingFlow(specificBundleState.flows, deposit);
        if (!matchingFlow) {
            throw new Error("Found bundle state containing flow but no matching flow found for deposit");
        }
        var lpFee = matchingFlow.lpFee.mul(deposit.amount).div(fixedPointAdjustment);
        var depositBalancingFee = matchingFlow.balancingFee;
        return {
            lpFee: lpFee,
            depositBalancingFee: depositBalancingFee,
            systemFee: lpFee.add(depositBalancingFee),
        };
    };
    /**
     * @notice Can be used by Relayer to approximate next refund balancing fees on refund chain. This is useful
     * for the relayer because they can only assume that their fill gets mined as the "next" fill on the destination
     * chain.
     * @dev Should work even if there are no validated flows or no bundle states yet.
     */
    UBAClientWithRefresh.prototype.computeBalancingFeeForNextRefund = function (repaymentChainId, refundTokenSymbol, amount) {
        this.assertUpdated();
        // Grab latest flow to get the latest running balance on the desired refund chain for the desired token.
        var latestBlockRange = this.ubaBundleBlockRanges.at(-1);
        if (!latestBlockRange) {
            throw new Error("No block ranges stored, have you updated this client?");
        }
        var lastBundleState = this.getBundleState(latestBlockRange, refundTokenSymbol, repaymentChainId);
        var latestRunningBalance = ethers.constants.Zero;
        var latestIncentiveBalance = ethers.constants.Zero;
        // Load latest bundle config.
        var ubaConfigForBundle = lastBundleState.ubaConfig;
        // Check if there are any flows in this bundle state. If there are not, then we can assume that
        // running balances are 0.
        var lastFlow = lastBundleState.flows.at(-1);
        if (isDefined(lastFlow)) {
            // Compute the fees that would be charged if `amount` was filled by the caller and the fill was
            // slotted in next on this chain.
            var _a = analog.calculateHistoricalRunningBalance(lastBundleState.flows.map(function (_a) {
                var flow = _a.flow, balancingFee = _a.balancingFee;
                return __assign(__assign({}, flow), { incentiveFee: balancingFee });
            }), lastFlow.runningBalance, lastFlow.incentiveBalance, lastFlow.netRunningBalanceAdjustment, repaymentChainId, refundTokenSymbol, ubaConfigForBundle), runningBalance = _a.runningBalance, incentiveBalance = _a.incentiveBalance;
            latestRunningBalance = runningBalance;
            latestIncentiveBalance = incentiveBalance;
        }
        var balancingFee = getRefundFee(amount, latestRunningBalance, latestIncentiveBalance, repaymentChainId, ubaConfigForBundle).balancingFee;
        return balancingFee;
    };
    /**
     * @notice Expected to be called by dataworker to reconstruct bundle roots using validated flow information
     * in bundle.
     */
    UBAClientWithRefresh.prototype.getModifiedFlows = function (chainId, tokenSymbol, startBlock, endBlock) {
        var _this = this;
        this.assertUpdated();
        // Find the bundle block range that entirely contains start and end block
        var specificBundleRange = this.ubaBundleBlockRanges.find(function (blockRanges) {
            var blockRangeForChain = getBlockRangeForChain(blockRanges, chainId, _this.chainIdIndices);
            return blockRangeForChain[0] <= startBlock && blockRangeForChain[1] >= endBlock;
        });
        if (!isDefined(specificBundleRange)) {
            throw new Error("Could not find bundle block range containing block range ".concat(startBlock, " to ").concat(endBlock, " on chain ").concat(chainId));
        }
        var bundleState = this.getBundleState(specificBundleRange, tokenSymbol, chainId);
        return bundleState.flows.filter(function (_a) {
            var flow = _a.flow;
            flow.blockNumber <= endBlock && flow.blockNumber >= startBlock;
        });
    };
    /**
     * Can be used by middleware API to refresh the latest bundle state for a given token and chain. The API
     * can serve this information and make it very fast to then compute the next deposit and refund balancing fees.
     * @param tokenSymbol
     * @param chainId
     * @returns
     */
    UBAClientWithRefresh.prototype.getLatestBundleState = function (tokenSymbol, chainId) {
        this.assertUpdated();
        var latestBlockRange = this.ubaBundleBlockRanges.at(-1);
        if (!latestBlockRange) {
            throw new Error("No block ranges stored, have you updated this client?");
        }
        var _a = this.getBundleState(latestBlockRange, tokenSymbol, chainId), flows = _a.flows, ubaConfig = _a.ubaConfig;
        return {
            flows: flows,
            ubaConfig: ubaConfig,
        };
    };
    // /////////////////
    //
    //
    // PRIVATE METHODS:
    //
    // /////////////////
    /**
     * This client should only access the `ubaBundleStates` state via this function.
     */
    UBAClientWithRefresh.prototype.getBundleState = function (blockRange, tokenSymbol, chainId) {
        var key = this.getKeyForBundle(blockRange, tokenSymbol, chainId);
        var bundleState = this.ubaBundleStates[key];
        if (!isDefined(bundleState)) {
            throw new Error("Bundle state for chain ".concat(chainId, ", token ").concat(tokenSymbol, " and block range ").concat(blockRange, " should exist"));
        }
        return bundleState;
    };
    /**
     * @notice Compute the LP fee for a given amount.
     */
    UBAClientWithRefresh.prototype.computeLpFee = function (baselineFee) {
        // @dev Temporarily, the LP fee only comprises the baselineFee. In the future, a variable component will be
        // added to the baseline fee that takes into account the utilized liquidity in the system and how the the bridge
        // defined by { amount, originChain, refundChain, hubPoolBlock } affects that liquidity.
        return baselineFee;
    };
    /**
     * Validate flows in bundle block ranges. Returns the flows mapped by chain ID where the flow occurred and also
     * caches the validated flows in this class instance.
     * @param blockRanges The block ranges to validate flows for. Must be a known block range in this.ubaBundleBlockRanges class variable.
     * @param tokenSymbol Flows to filter on.
     * @returns
     */
    UBAClientWithRefresh.prototype.validateFlowsInBundle = function (blockRanges, tokenSymbol) {
        var _a;
        return __awaiter(this, void 0, void 0, function () {
            var l1TokenAddress, blockRangesForChain, flowsInBundle, _b, _i, flowsInBundle_1, flow, flowChain, validatedFlow;
            var _this = this;
            return __generator(this, function (_c) {
                switch (_c.label) {
                    case 0:
                        // This should only be called with exact block ranges already stored in this.ubaBundleBlockRanges.
                        if (!this.ubaBundleBlockRanges.some(function (bundleBlockRanges) {
                            return JSON.stringify(blockRanges) === JSON.stringify(bundleBlockRanges);
                        })) {
                            throw new Error("Invalid block ranges not found in this.ubaBundleBlockRanges state");
                        }
                        l1TokenAddress = (_a = this.hubPoolClient.getL1Tokens().find(function (token) { return token.symbol === tokenSymbol; })) === null || _a === void 0 ? void 0 : _a.address;
                        if (!l1TokenAddress)
                            throw new Error("No L1 token address found for token symbol ".concat(tokenSymbol));
                        // Make sure our spoke pool clients have a large enough block range to cover this range.
                        if (blockRangesAreInvalidForSpokeClients(this.spokePoolClients, blockRanges, this.chainIdIndices)) {
                            throw new Error("Spoke pool clients do not have the block ranges necessary to look up data for bundle ranges: ".concat(JSON.stringify(blockRanges)));
                        }
                        blockRangesForChain = Object.fromEntries(this.enabledChainIds.map(function (chainId) {
                            return [chainId, getBlockRangeForChain(blockRanges, chainId, _this.chainIdIndices)];
                        }));
                        _b = sortFlowsAscending;
                        return [4 /*yield*/, mapAsync(this.enabledChainIds, function (chainId) { return __awaiter(_this, void 0, void 0, function () {
                                var _a, startBlock, endBlock, flows;
                                return __generator(this, function (_b) {
                                    switch (_b.label) {
                                        case 0:
                                            _a = blockRangesForChain[chainId], startBlock = _a[0], endBlock = _a[1];
                                            // Don't load flows for disabled block ranges in this bundle.
                                            if (startBlock === endBlock) {
                                                return [2 /*return*/, []];
                                            }
                                            return [4 /*yield*/, getUBAFlows(tokenSymbol, chainId, this.spokePoolClients, this.hubPoolClient, startBlock, endBlock)];
                                        case 1:
                                            flows = _b.sent();
                                            // // Print out readable breakdown of flows.
                                            // const prettyFlows = flows.reduce(
                                            //   (acc, flow) => {
                                            //     if (isUbaOutflow(flow) && outflowIsFill(flow)) {
                                            //       acc.fills += 1;
                                            //     } else if (isUbaOutflow(flow) && outflowIsRefund(flow)) {
                                            //       acc.refunds += 1;
                                            //     } else {
                                            //       acc.deposits += 1;
                                            //     }
                                            //     return acc;
                                            //   },
                                            //   { fills: 0, refunds: 0, deposits: 0 }
                                            // );
                                            // this.logger.debug({
                                            //   at: "UBAClientWithRefresh#validateFlowsInBundle",
                                            //   message: `Flow breakdown for chain ${chainId} and token ${tokenSymbol}`,
                                            //   blockRangesForChain,
                                            //   prettyFlows
                                            // })
                                            return [2 /*return*/, flows];
                                    }
                                });
                            }); })];
                    case 1:
                        flowsInBundle = _b.apply(void 0, [(_c.sent()).flat()]);
                        _i = 0, flowsInBundle_1 = flowsInBundle;
                        _c.label = 2;
                    case 2:
                        if (!(_i < flowsInBundle_1.length)) return [3 /*break*/, 5];
                        flow = flowsInBundle_1[_i];
                        flowChain = getFlowChain(flow);
                        return [4 /*yield*/, this.validateFlow(flow)];
                    case 3:
                        validatedFlow = _c.sent();
                        if (isDefined(validatedFlow)) {
                            // console.log("Validated ✅", {
                            //   runningBalance: validatedFlow.runningBalance,
                            //   incentiveBalance: validatedFlow.incentiveBalance.toString(),
                            //   netRunningBalanceAdjustment: validatedFlow.netRunningBalanceAdjustment.toString(),
                            //   balancingFee: validatedFlow.balancingFee.toString(),
                            //   lpFee: validatedFlow.lpFee.toString(),
                            // });
                            this.appendValidatedFlowsToClassState(flowChain, tokenSymbol, [validatedFlow], blockRanges);
                        }
                        else {
                            // this.logger.debug({
                            //   at: "UBAClientWithRefresh#validateFlowsInBundle",
                            //   message: `❌ Invalidated flow for chain ${flowChain} and token ${tokenSymbol}`,
                            //   blockRanges: blockRangesForChain,
                            //   flow
                            // })
                        }
                        _c.label = 4;
                    case 4:
                        _i++;
                        return [3 /*break*/, 2];
                    case 5: return [2 /*return*/, Object.fromEntries(this.enabledChainIds.map(function (chainId) {
                            var bundleStateForChain = _this.getBundleState(blockRanges, tokenSymbol, chainId);
                            return [chainId, bundleStateForChain.flows];
                        }))];
                }
            });
        });
    };
    /**
     * Return flow with computed fees if it is valid. Otherwise, return undefined. Inflows are always valid,
     * while outflows need to match against an inflow and have set the correct realizedLpFeePct
     * @param flow Flow to validate. It's assumed that flow is already matched with a deposit, if its an outflow.
     * @returns Validated flow or undefined if flow cannot be validated.
     */
    UBAClientWithRefresh.prototype.validateFlow = function (flow) {
        var _a, _b, _c, _d, _e, _f, _g;
        return __awaiter(this, void 0, void 0, function () {
            var flowChain, tokenSymbol, l1TokenAddress, blockRangesContainingFlow, bundleSharedState, openingBalanceForChain, ubaConfigForChain, precedingValidatedFlows, lastFlow, latestRunningBalance, latestIncentiveBalance, latestNetRunningBalanceAdjustment, potentialBalancingFee, _h, runningBalance, incentiveBalance, netRunningBalanceAdjustment, lpFeePct, lpFee, newModifiedFlow, matchedDeposit, bundleForMatchedDeposit, depositBundleState, ubaConfigForDeposit, lpFeePct_1, matchedDepositFlow, matchedDepositBlockRangeInFlowBundle, matchedDepositBundleFlows, validatedFlowsOnOriginChain, matchedDepositBundleFlows, expectedRealizedLpFeeForMatchedDeposit, expectedRealizedLpFeePctForMatchedDeposit;
            return __generator(this, function (_j) {
                switch (_j.label) {
                    case 0:
                        if (isUbaInflow(flow)) {
                            flowChain = flow.originChainId;
                            tokenSymbol = (_a = this.hubPoolClient.getL1TokenInfoForL2Token(flow.originToken, flowChain)) === null || _a === void 0 ? void 0 : _a.symbol;
                        }
                        else {
                            // If the flow is a fill, then we need to validate its matched deposit.
                            if (outflowIsFill(flow)) {
                                flowChain = flow.destinationChainId;
                                tokenSymbol = (_b = this.hubPoolClient.getL1TokenInfoForL2Token(flow.destinationToken, flowChain)) === null || _b === void 0 ? void 0 : _b.symbol;
                            }
                            else {
                                flowChain = flow.repaymentChainId;
                                tokenSymbol = (_c = this.hubPoolClient.getL1TokenInfoForL2Token(flow.refundToken, flowChain)) === null || _c === void 0 ? void 0 : _c.symbol;
                            }
                        }
                        if (!isDefined(tokenSymbol))
                            throw new Error("No token symbol found");
                        l1TokenAddress = (_d = this.hubPoolClient.getL1Tokens().find(function (token) { return token.symbol === tokenSymbol; })) === null || _d === void 0 ? void 0 : _d.address;
                        if (!isDefined(l1TokenAddress))
                            throw new Error("No L1 token address found for token symbol");
                        blockRangesContainingFlow = this.getUbaBundleBlockRangeContainingFlow(flow.blockNumber, flowChain);
                        bundleSharedState = this.getBundleState(blockRangesContainingFlow, tokenSymbol, flowChain);
                        openingBalanceForChain = bundleSharedState.openingBalances;
                        ubaConfigForChain = bundleSharedState.ubaConfig;
                        precedingValidatedFlows = bundleSharedState.flows;
                        lastFlow = precedingValidatedFlows[precedingValidatedFlows.length - 1];
                        latestRunningBalance = (_e = lastFlow === null || lastFlow === void 0 ? void 0 : lastFlow.runningBalance) !== null && _e !== void 0 ? _e : openingBalanceForChain.runningBalance;
                        latestIncentiveBalance = (_f = lastFlow === null || lastFlow === void 0 ? void 0 : lastFlow.incentiveBalance) !== null && _f !== void 0 ? _f : openingBalanceForChain.incentiveBalance;
                        latestNetRunningBalanceAdjustment = (_g = lastFlow === null || lastFlow === void 0 ? void 0 : lastFlow.netRunningBalanceAdjustment) !== null && _g !== void 0 ? _g : ethers.constants.Zero;
                        if (isUbaInflow(flow)) {
                            (potentialBalancingFee = getDepositFee(flow.amount, latestRunningBalance, latestIncentiveBalance, flowChain, ubaConfigForChain).balancingFee);
                        }
                        else {
                            (potentialBalancingFee = getRefundFee(flow.amount, latestRunningBalance, latestIncentiveBalance, flowChain, ubaConfigForChain).balancingFee);
                        }
                        _h = analog.calculateHistoricalRunningBalance([
                            __assign(__assign({}, flow), { incentiveFee: potentialBalancingFee }),
                        ], latestRunningBalance, latestIncentiveBalance, latestNetRunningBalanceAdjustment, flowChain, tokenSymbol, ubaConfigForChain), runningBalance = _h.runningBalance, incentiveBalance = _h.incentiveBalance, netRunningBalanceAdjustment = _h.netRunningBalanceAdjustment;
                        lpFeePct = this.computeLpFee(ubaConfigForChain.getBaselineFee(flow.destinationChainId, flow.originChainId));
                        lpFee = lpFeePct.mul(flow.amount).div(fixedPointAdjustment);
                        newModifiedFlow = {
                            flow: flow,
                            balancingFee: potentialBalancingFee,
                            lpFee: lpFee,
                            runningBalance: runningBalance,
                            incentiveBalance: incentiveBalance,
                            netRunningBalanceAdjustment: netRunningBalanceAdjustment,
                        };
                        if (!isUbaInflow(flow)) return [3 /*break*/, 1];
                        return [2 /*return*/, newModifiedFlow];
                    case 1:
                        // Now we need to validate the refund or fill.
                        if (!isDefined(flow.realizedLpFeePct)) {
                            throw new Error("Outflow has undefined realizedLpFeePct");
                        }
                        // ASSUMPTION: the flow is already matched against a deposit by `getUBAFlows` when comparing
                        // all params besides `realizedLpFeePct`.
                        // We need to make sure that the matched deposit is not a pre UBA deposit. Only pre UBA deposits
                        // have defined realizedLpFeePct's at the time they are loaded in the flow.matchedDeposit
                        // entry at this stage so they are trivial to validate.
                        if (isDefined(flow.matchedDeposit.realizedLpFeePct)) {
                            // console.log("- Flow matched a pre-UBA deposit");
                            // TODO: Potentially add additional safety check that we're identifying the matched deposit correctly as
                            // a pre UBA deposit. If the deposit is not pre UBA but has a set realizedLpFeePct at this point
                            // its a bug.
                            if (flow.matchedDeposit.realizedLpFeePct.eq(flow.realizedLpFeePct)) {
                                // If flow matched with a pre UBA deposit then the flow should accrue no balancing fees and not impact
                                // running balances so we should use the running balances from before the flow.
                                return [2 /*return*/, {
                                        flow: flow,
                                        runningBalance: latestRunningBalance,
                                        incentiveBalance: latestIncentiveBalance,
                                        netRunningBalanceAdjustment: latestNetRunningBalanceAdjustment,
                                        // Balancing fee for a pre UBA refund is 0
                                        balancingFee: ethers.constants.Zero,
                                        // Set realized LP fee for fill equal to the realized LP fee for the matched deposit.
                                        lpFee: flow.realizedLpFeePct.mul(flow.amount).div(fixedPointAdjustment),
                                    }];
                            }
                            else {
                                this.logger.debug({
                                    at: "UBAClientWithRefresh#validateFlow",
                                    message: "Flow matched a pre-UBA deposit, but it was invalid because it set the wrong LP fee",
                                    flow: flow,
                                });
                                return [2 /*return*/, undefined];
                            }
                        }
                        matchedDeposit = flow.matchedDeposit;
                        bundleForMatchedDeposit = this.getUbaBundleBlockRangeContainingFlow(matchedDeposit.blockNumber, matchedDeposit.originChainId);
                        depositBundleState = this.getBundleState(bundleForMatchedDeposit, tokenSymbol, matchedDeposit.originChainId);
                        ubaConfigForDeposit = depositBundleState.ubaConfig;
                        if (ubaConfigForDeposit.isBalancingFeeCurveFlatAtZero(flow.matchedDeposit.originChainId)) {
                            lpFeePct_1 = newModifiedFlow.lpFee.mul(fixedPointAdjustment).div(flow.amount);
                            if (!lpFeePct_1.eq(flow.realizedLpFeePct)) {
                                // this.logger.debug({
                                //   at: "UBAClientWithRefresh#validateFlow",
                                //   message: `Flow matched a deposit on a chain with a flat balancing fee curve at 0, but it was invalid because it set the wrong LP fee`,
                                //   expected: lpFeePct.toString(),
                                //   flow
                                // })
                                return [2 /*return*/, undefined];
                            }
                            else {
                                return [2 /*return*/, newModifiedFlow];
                            }
                        }
                        if (!(flow.blockTimestamp < flow.matchedDeposit.blockTimestamp)) return [3 /*break*/, 2];
                        // We cannot invalidate a fill if the latest block.timestamp on the deposit.origin chain is not > than the
                        // the fill's timestamp. This is because its still possible to send a deposit on the origin chain
                        // that would validate this fill.
                        // TODO: Figure out how to handle this without crashing.
                        if (this.latestBlockTimestamps[flow.matchedDeposit.originChainId] < flow.blockTimestamp) {
                            this.logger.error({
                                at: "UBAClientWithRefresh#validateFlow",
                                message: "We cannot invalidate a fill if the latest block.timestamp on the origin chain is <= the fill's timestamp",
                                latestOriginChainBlockTimestamp: this.latestBlockTimestamps[flow.matchedDeposit.originChainId],
                                flow: flow,
                            });
                            throw new Error("We cannot invalidate a fill if the latest block.timestamp ".concat(this.latestBlockTimestamps[flow.matchedDeposit.originChainId], " on the deposit.origin chain is not > than the fill's timestamp ").concat(flow.blockTimestamp, ". This is because its still possible to send a deposit on the origin chain that would validate this fill."));
                        }
                        return [3 /*break*/, 10];
                    case 2:
                        matchedDepositFlow = void 0;
                        matchedDepositBlockRangeInFlowBundle = getBlockRangeForChain(blockRangesContainingFlow, matchedDeposit.originChainId, this.chainIdIndices);
                        if (!(matchedDeposit.blockNumber > matchedDepositBlockRangeInFlowBundle[1])) return [3 /*break*/, 6];
                        // Its possible that a matched deposit is in a later bundle because of the way the dataworker constructs
                        // bundle block ranges. This is possible if the fill is near the end of the previous bundle before the deposit
                        // which is at the beginning.
                        // We first try to check existing validated flows in the later bundle to see if we've already
                        // validated this flow.
                        // console.log("- ➡️ Matched deposit bundle block range is after flow's", {
                        //   bundleForMatchedDeposit,
                        // });
                        matchedDepositFlow = this.getMatchingValidatedFlow(matchedDeposit.blockNumber, matchedDeposit.originChainId, matchedDeposit, tokenSymbol);
                        if (!isDefined(matchedDepositFlow)) return [3 /*break*/, 3];
                        return [3 /*break*/, 5];
                    case 3: return [4 /*yield*/, this.validateFlowsInBundle(bundleForMatchedDeposit, tokenSymbol)];
                    case 4:
                        matchedDepositBundleFlows = _j.sent();
                        matchedDepositFlow = getMatchingFlow(matchedDepositBundleFlows[matchedDeposit.originChainId], matchedDeposit);
                        _j.label = 5;
                    case 5: return [3 /*break*/, 9];
                    case 6:
                        if (!(matchedDeposit.blockNumber >= matchedDepositBlockRangeInFlowBundle[0] &&
                            matchedDeposit.blockNumber <= matchedDepositBlockRangeInFlowBundle[1])) return [3 /*break*/, 7];
                        validatedFlowsOnOriginChain = depositBundleState.flows;
                        matchedDepositFlow = getMatchingFlow(validatedFlowsOnOriginChain, matchedDeposit);
                        if (!isDefined(matchedDepositFlow)) {
                            this.logger.error({
                                at: "UBAClientWithRefresh#validateFlow",
                                message: "Deposit should be in same bundle as flow and already have been validated but cannot be found",
                                validatedFlowsOnOriginChain: validatedFlowsOnOriginChain,
                                flow: flow,
                            });
                            throw new Error("Could not find matched deposit in same bundle as fill");
                        }
                        return [3 /*break*/, 9];
                    case 7:
                        // The bundle containing the matched deposit is older than current bundle range for flow.
                        // We might need to recurse here if we haven't validated the deposit yet.
                        // console.log("- ⬅️ Matched deposit bundle block range is older than flow's", {
                        //   bundleForMatchedDeposit,
                        // });
                        matchedDepositFlow = this.getMatchingValidatedFlow(matchedDeposit.blockNumber, matchedDeposit.originChainId, matchedDeposit, tokenSymbol);
                        if (!!isDefined(matchedDepositFlow)) return [3 /*break*/, 9];
                        return [4 /*yield*/, this.validateFlowsInBundle(bundleForMatchedDeposit, tokenSymbol)];
                    case 8:
                        matchedDepositBundleFlows = _j.sent();
                        matchedDepositFlow = getMatchingFlow(matchedDepositBundleFlows[matchedDeposit.originChainId], matchedDeposit);
                        _j.label = 9;
                    case 9:
                        // At this point we couldn't find a flow in the cache or from newly validated data that matches the
                        // flow.matchedDeposit. This is unexpected and perhaps means we need to widen the spoke pool client lookback.
                        if (!isDefined(matchedDepositFlow)) {
                            this.logger.error({
                                at: "UBAClientWithRefresh#validateFlow",
                                message: "No matching flow found in client state for matched deposit",
                                flow: flow,
                            });
                            throw new Error("Could not validate or invalidate matched deposit");
                        }
                        expectedRealizedLpFeeForMatchedDeposit = matchedDepositFlow.lpFee.add(matchedDepositFlow.balancingFee);
                        expectedRealizedLpFeePctForMatchedDeposit = expectedRealizedLpFeeForMatchedDeposit
                            .mul(fixedPointAdjustment)
                            .div(matchedDepositFlow.flow.amount);
                        // console.log(
                        //   `- Expected realized lp fee pct for matched deposit: ${expectedRealizedLpFeePctForMatchedDeposit.toString()}, actual: ${flow.realizedLpFeePct?.toString()}`
                        // );
                        // We allow for some precision loss because of the way we compute balancing fees. The balancing fee curves are integrated
                        // by computePiecewiseLinearFunction and they return values, so we set balancing fees equal to those values divided by
                        // flow amounts, which produces precision loss.
                        if (expectedRealizedLpFeePctForMatchedDeposit.lte(flow.realizedLpFeePct.mul(toBNWei("1.01")).div(fixedPointAdjustment)) ||
                            expectedRealizedLpFeePctForMatchedDeposit.gte(flow.realizedLpFeePct.mul(toBNWei("0.99")).div(fixedPointAdjustment))) {
                            return [2 /*return*/, newModifiedFlow];
                        }
                        else {
                            // this.logger.debug({
                            //   at: "UBAClientWithRefresh#validateFlow",
                            //   message: `Matched deposit was invalidated by incorrect realized lp fee pct set for outflow`,
                            //   expected: expectedRealizedLpFeePctForMatchedDeposit.toString(),
                            //   matchedDepositFlow,
                            //   flow,
                            // })
                        }
                        _j.label = 10;
                    case 10: return [2 /*return*/, undefined];
                }
            });
        });
    };
    /**
     * Locates the bundle range containing the flow on the specified chain and then tries to return the matching
     * validated flow if it already exists in memory.
     * @param flowBlock
     * @param flowChain
     * @param flow
     * @returns
     */
    UBAClientWithRefresh.prototype.getMatchingValidatedFlow = function (flowBlock, flowChain, flow, tokenSymbol) {
        var bundleContainingFlow = this.getUbaBundleBlockRangeContainingFlow(flowBlock, flowChain);
        var validatedFlowsInBundle = this.getBundleState(bundleContainingFlow, tokenSymbol, flowChain).flows;
        var matchingFlow = getMatchingFlow(validatedFlowsInBundle, flow);
        return matchingFlow;
    };
    /**
     * Returns a convenient mapping of the most recent block ranges for each chain.
     * Assumption: For each chain, there should be the same number of block ranges returned.
     * Assumption: The block ranges for each chain should cover from the activation bundle start block
     * for the UBA until the latest block searched for the chain.
     * @param bundleCount Block ranges to fetch per chain
     * @returns A dictionary mapping chainId to an array of block ranges, where the ranges are arrays of length two
     *         and the first element is the start block and the second element is the end block.
     */
    UBAClientWithRefresh.prototype.getMostRecentBundleBlockRangesPerChain = function (bundleCount) {
        var _this = this;
        return this.chainIdIndices.reduce(function (acc, chainId) {
            var _a;
            // Gets the most recent `bundleCount` block ranges for this chain.
            var _blockRangesForChain = getMostRecentBundleBlockRanges(chainId, bundleCount, _this.hubPoolClient, _this.spokePoolClients).map(function (_a) {
                var start = _a.start, end = _a.end;
                return [start, end];
            });
            // Sanity check that block ranges cover from UBA activation bundle start block for chain to latest spoke pool
            // client block searched:
            var ubaActivationBundleStartBlockForChain = getBlockForChain(getUbaActivationBundleStartBlocks(_this.hubPoolClient), chainId, _this.chainIdIndices);
            if (_blockRangesForChain.length === 0) {
                throw new Error("Should never return 0 length block ranges for chain ".concat(chainId));
            }
            if (
            // Check 1: start block of first block range should be equal to UBA activation bundle start block for chain
            _blockRangesForChain[0][0] !== ubaActivationBundleStartBlockForChain ||
                // Check 2: end block of last block range should be equal to latest spoke pool client block searched
                (isDefined(_this.spokePoolClients[chainId]) &&
                    // eslint-disable-next-line @typescript-eslint/no-non-null-assertion
                    _blockRangesForChain.at(-1)[1] !== _this.spokePoolClients[chainId].latestBlockSearched)) {
                _this.logger.error({
                    at: "UBAClientWithRefresh#getMostRecentBundleBlockRangesPerChain",
                    message: "Block ranges for chain ".concat(chainId, " do not cover from UBA activation bundle start block to latest spoke pool client block searched"),
                    startBlockForChain: _blockRangesForChain[0][0],
                    ubaActivationBundleStartBlockForChain: ubaActivationBundleStartBlockForChain,
                    endBlockForChain: (_a = _blockRangesForChain.at(-1)) === null || _a === void 0 ? void 0 : _a[1],
                    latestSpokePoolClientBlockSearched: _this.spokePoolClients[chainId].latestBlockSearched,
                });
                throw new Error("Block ranges for chain ".concat(chainId, " do not cover from UBA activation bundle start block to latest spoke pool client block searched"));
            }
            // Map the block ranges to this chain and move on to the next chain.
            acc[chainId] = _blockRangesForChain;
            return acc;
        }, {});
    };
    /**
     * Return the block range stored in this.ubaBundleBlockRanges that contains the flow.
     * @param flowBlock
     * @param flowChain
     * @returns
     */
    UBAClientWithRefresh.prototype.getUbaBundleBlockRangeContainingFlow = function (flowBlock, flowChain) {
        var _this = this;
        var blockRangesContainingFlow = this.ubaBundleBlockRanges.find(function (blockRanges) {
            var blockRangeForChain = getBlockRangeForChain(blockRanges, flowChain, _this.chainIdIndices);
            return blockRangeForChain[0] <= flowBlock && flowBlock <= blockRangeForChain[1];
        });
        if (!blockRangesContainingFlow) {
            throw new Error("Could not find bundle block range containing flow at block ".concat(flowBlock, " on chain ").concat(flowChain));
        }
        return blockRangesContainingFlow;
    };
    /**
     * This client should only update the `ubaBundleStates` state via this function.
     */
    UBAClientWithRefresh.prototype.appendValidatedFlowsToClassState = function (chainId, tokenSymbol, flows, bundleBlockRanges) {
        var _a;
        var bundleState = this.getBundleState(bundleBlockRanges, tokenSymbol, chainId);
        (_a = bundleState.flows).push.apply(_a, flows);
    };
    /**
     * Updates the bundle state.
     */
    UBAClientWithRefresh.prototype.update = function () {
        return __awaiter(this, void 0, void 0, function () {
            var latestHubPoolBlock, tokens, blockRangesByChain, bundleBlockRangesCount, bundleBlockRanges, _loop_1, this_1, i, _loop_2, this_2, i, latestTimestampsPerChain, _loop_3, this_3, i, newUbaClientState, _loop_4, this_4, i, _loop_5, this_5, i, _loop_6, i;
            var _this = this;
            return __generator(this, function (_a) {
                switch (_a.label) {
                    case 0:
                        latestHubPoolBlock = this.hubPoolClient.latestBlockSearched;
                        this.logger.debug({
                            at: "UBAClientWithRefresh",
                            message: "❤️‍🔥😭  Updating UBA Client",
                            enabledChains: this.enabledChainIds,
                        });
                        tokens = this.tokens;
                        // Load all UBA bundle block ranges for each chain:
                        this.logger.debug({
                            at: "UBAClientWithRefresh#getMostRecentBundleBlockRangesPerChain",
                            message: "Loaded UBA bundle start blocks",
                            ubaActivationBundleStartBlockForChain: getUbaActivationBundleStartBlocks(this.hubPoolClient),
                        });
                        blockRangesByChain = this.getMostRecentBundleBlockRangesPerChain(100);
                        bundleBlockRangesCount = blockRangesByChain[this.chainIdIndices[0]].length;
                        bundleBlockRanges = [];
                        _loop_1 = function (i) {
                            bundleBlockRanges.push(this_1.chainIdIndices
                                .map(function (chainId) {
                                // If chain has a bundle beginning at chain 0, then the chain was recently
                                // added to the chain ID list, so fill preceding block ranges with zero length ranges so that
                                // the number of block ranges for this chain equals the number of block ranges for other chains.
                                var blockRangeCountForChain = blockRangesByChain[chainId].length;
                                if (blockRangeCountForChain < bundleBlockRangesCount) {
                                    var firstBlockRange = blockRangesByChain[chainId][0];
                                    // Add block ranges with zero length to the beginning of the bundle block ranges for this chain
                                    if (i < bundleBlockRangesCount - blockRangeCountForChain) {
                                        return [firstBlockRange[0], firstBlockRange[0]];
                                    }
                                    else {
                                        return firstBlockRange;
                                    }
                                }
                                return blockRangesByChain[chainId][i];
                            })
                                .filter(isDefined));
                        };
                        this_1 = this;
                        for (i = 0; i < bundleBlockRangesCount; i++) {
                            _loop_1(i);
                        }
                        // Validate that block ranges are equal length for all bundles.
                        if (bundleBlockRanges.some(function (blockRange) { return blockRange.length !== _this.chainIdIndices.length; })) {
                            throw new Error("Block ranges are not equal length for all bundles");
                        }
                        this.ubaBundleBlockRanges = bundleBlockRanges;
                        this.logger.debug({
                            at: "UBAClientWithRefresh#update",
                            message: "UBA bundle block ranges we're storing in class memory",
                            bundleBlockRanges: this.ubaBundleBlockRanges,
                        });
                        _loop_2 = function (i) {
                            var bundleBlockRange = this_2.ubaBundleBlockRanges[i];
                            var startBlocks = bundleBlockRange.map(function (_a) {
                                var start = _a[0];
                                return start;
                            });
                            var mainnetStartBlock = getBlockForChain(startBlocks, this_2.hubPoolClient.chainId, this_2.chainIdIndices);
                            tokens.forEach(function (token) {
                                var _a;
                                var l1TokenAddress = (_a = _this.hubPoolClient.getL1Tokens().find(function (_token) { return _token.symbol === token; })) === null || _a === void 0 ? void 0 : _a.address;
                                if (!isDefined(l1TokenAddress))
                                    throw new Error("No L1 token address found for token symbol ".concat(token));
                                _this.enabledChainIds.forEach(function (chainId) {
                                    var bundleStateKey = _this.getKeyForBundle(bundleBlockRange, token, chainId);
                                    var startBlock = getBlockForChain(startBlocks, chainId, _this.chainIdIndices);
                                    var openingBalances = getOpeningRunningBalanceForEvent(_this.hubPoolClient, 
                                    // Pass in a start block for the bundle containing the flow to this function so we always
                                    // get a running balance from the last validated bundle before this flow's bundle.
                                    startBlock, chainId, l1TokenAddress, latestHubPoolBlock);
                                    // console.log(
                                    //   `Setting running balance ${openingBalances.runningBalance.toString()} for bundle range ${bundleStateKey}}`
                                    // );
                                    var ubaConfig = getUBAFeeConfig(_this.hubPoolClient, chainId, token, mainnetStartBlock);
                                    _this.ubaBundleStates[bundleStateKey] = {
                                        openingBalances: openingBalances,
                                        ubaConfig: ubaConfig,
                                        flows: [],
                                        loadedFromCache: false,
                                    };
                                });
                            });
                        };
                        this_2 = this;
                        // Pre-load bundle configs and opening balances for each chain and token per bundle range.
                        // This also instantiates the ubaBundleStates class dictionary.
                        for (i = 0; i < this.ubaBundleBlockRanges.length; i++) {
                            _loop_2(i);
                        }
                        latestTimestampsPerChain = Object.fromEntries(this.enabledChainIds
                            .map(function (chainId) {
                            if (!isDefined(_this.spokePoolClients[chainId]))
                                return undefined;
                            return [chainId, _this.spokePoolClients[chainId].getCurrentTime()];
                        })
                            .filter(isDefined));
                        this.latestBlockTimestamps = latestTimestampsPerChain;
                        this.logger.debug({
                            at: "UBAClientWithRefresh#update",
                            message: "Latest block timestamps per chain",
                            latestBlockTimestamps: this.latestBlockTimestamps,
                        });
                        if (!isDefined(this.cachingMechanism)) return [3 /*break*/, 4];
                        _loop_3 = function (i) {
                            var mostRecentBundleBlockRanges, mainnetBundleStartBlock;
                            return __generator(this, function (_b) {
                                switch (_b.label) {
                                    case 0:
                                        mostRecentBundleBlockRanges = this_3.ubaBundleBlockRanges[i];
                                        mainnetBundleStartBlock = getBlockRangeForChain(mostRecentBundleBlockRanges, this_3.hubPoolClient.chainId, this_3.chainIdIndices)[0];
                                        if (isDefined(this_3.clientConfig.latestMainnetBundleStartBlockToLoadFromCache) &&
                                            this_3.clientConfig.latestMainnetBundleStartBlockToLoadFromCache > mainnetBundleStartBlock) {
                                            this_3.logger.debug({
                                                at: "UBAClientWithRefresh#update",
                                                message: "Skipping loading bundle data from cache for bundle with mainnet start block ".concat(mainnetBundleStartBlock, " because it's older than the oldest bundle we want to load from the cache"),
                                                bundleBlockRanges: mostRecentBundleBlockRanges,
                                                oldestMainnetStartBlockToLoadFromCache: this_3.clientConfig.latestMainnetBundleStartBlockToLoadFromCache,
                                            });
                                            return [2 /*return*/, "continue"];
                                        }
                                        return [4 /*yield*/, forEachAsync(tokens, function (token) { return __awaiter(_this, void 0, void 0, function () {
                                                var _this = this;
                                                return __generator(this, function (_a) {
                                                    switch (_a.label) {
                                                        case 0: return [4 /*yield*/, forEachAsync(this.enabledChainIds, function (chainId) { return __awaiter(_this, void 0, void 0, function () {
                                                                var redisKeyForBundle, modifiedFlowsInBundle;
                                                                var _a;
                                                                return __generator(this, function (_b) {
                                                                    switch (_b.label) {
                                                                        case 0:
                                                                            redisKeyForBundle = this.getKeyForBundle(mostRecentBundleBlockRanges, token, chainId);
                                                                            return [4 /*yield*/, ((_a = this.cachingMechanism) === null || _a === void 0 ? void 0 : _a.get(redisKeyForBundle))];
                                                                        case 1:
                                                                            modifiedFlowsInBundle = _b.sent();
                                                                            if (isDefined(modifiedFlowsInBundle)) {
                                                                                this.logger.debug({
                                                                                    at: "UBAClientWithRefresh#update",
                                                                                    message: "\uD83D\uDCBF Loaded bundle state from cache using key ".concat(redisKeyForBundle),
                                                                                    bundleBlockRanges: mostRecentBundleBlockRanges,
                                                                                    flowsLoaded: modifiedFlowsInBundle.length,
                                                                                });
                                                                                this.appendValidatedFlowsToClassState(chainId, token, modifiedFlowsInBundle, mostRecentBundleBlockRanges);
                                                                                // Mark as loaded from cache.
                                                                                this.getBundleState(mostRecentBundleBlockRanges, token, chainId).loadedFromCache = true;
                                                                            }
                                                                            else {
                                                                                this.logger.debug({
                                                                                    at: "UBAClientWithRefresh#update",
                                                                                    message: "No entry for key ".concat(redisKeyForBundle, " in redis"),
                                                                                    bundleBlockRanges: mostRecentBundleBlockRanges,
                                                                                });
                                                                            }
                                                                            return [2 /*return*/];
                                                                    }
                                                                });
                                                            }); })];
                                                        case 1:
                                                            _a.sent();
                                                            return [2 /*return*/];
                                                    }
                                                });
                                            }); })];
                                    case 1:
                                        _b.sent();
                                        return [2 /*return*/];
                                }
                            });
                        };
                        this_3 = this;
                        i = this.ubaBundleBlockRanges.length - 1;
                        _a.label = 1;
                    case 1:
                        if (!(i >= 0)) return [3 /*break*/, 4];
                        return [5 /*yield**/, _loop_3(i)];
                    case 2:
                        _a.sent();
                        _a.label = 3;
                    case 3:
                        i--;
                        return [3 /*break*/, 1];
                    case 4:
                        newUbaClientState = {};
                        _loop_4 = function (i) {
                            var mostRecentBundleBlockRanges;
                            return __generator(this, function (_c) {
                                switch (_c.label) {
                                    case 0:
                                        mostRecentBundleBlockRanges = this_4.ubaBundleBlockRanges[i];
                                        // console.log("Validating flows for bundle", mostRecentBundleBlockRanges);
                                        return [4 /*yield*/, forEachAsync(tokens, function (token) { return __awaiter(_this, void 0, void 0, function () {
                                                var modifiedFlowsInBundle;
                                                var _this = this;
                                                return __generator(this, function (_a) {
                                                    switch (_a.label) {
                                                        case 0:
                                                            if (!this.enabledChainIds.every(function (chainId) {
                                                                return _this.getBundleState(mostRecentBundleBlockRanges, token, chainId).loadedFromCache;
                                                            })) return [3 /*break*/, 1];
                                                            // const bundleKeyForBlockRanges = getBundleKeyForBlockRanges(mostRecentBundleBlockRanges);
                                                            // console.log(
                                                            //   `- Skipping validation for bundle ${bundleKeyForBlockRanges} for token ${token} because flows for all chains are already cached`
                                                            // );
                                                            modifiedFlowsInBundle = Object.fromEntries(this.enabledChainIds.map(function (chainId) {
                                                                var cachedFlows = _this.getBundleState(mostRecentBundleBlockRanges, token, chainId).flows;
                                                                return [chainId, cachedFlows];
                                                            }));
                                                            return [3 /*break*/, 3];
                                                        case 1: return [4 /*yield*/, this.validateFlowsInBundle(mostRecentBundleBlockRanges, token)];
                                                        case 2:
                                                            // Validate flows, which should load them into memory. This is a big expensive function especially
                                                            // if we haven't loaded any data from the cache and a lot of time has passed since the UBA activation
                                                            // bundle!
                                                            modifiedFlowsInBundle = _a.sent();
                                                            _a.label = 3;
                                                        case 3: 
                                                        // Save into UBA client state
                                                        return [4 /*yield*/, forEachAsync(this.enabledChainIds, function (chainId) { return __awaiter(_this, void 0, void 0, function () {
                                                                var bundleState, redisKeyForBundle;
                                                                return __generator(this, function (_a) {
                                                                    switch (_a.label) {
                                                                        case 0:
                                                                            if (!isDefined(newUbaClientState[chainId]))
                                                                                newUbaClientState[chainId] = {};
                                                                            if (!isDefined(newUbaClientState[chainId][token]))
                                                                                newUbaClientState[chainId][token] = [];
                                                                            bundleState = this.getBundleState(mostRecentBundleBlockRanges, token, chainId);
                                                                            newUbaClientState[chainId][token].push({
                                                                                bundleBlockRanges: mostRecentBundleBlockRanges,
                                                                                flows: bundleState.flows,
                                                                                ubaConfig: bundleState.ubaConfig,
                                                                                openingBalances: bundleState.openingBalances,
                                                                            });
                                                                            // We shouldn't cache the latest bundle state as it will always be unexecuted (i.e. its still
                                                                            // pending the challenge period or its leaves haven't been fully executed).
                                                                            // Moreover, we shouldn't cache bundles until we've seen a minimum number of bundles executed since UBA
                                                                            // genesis for safety reasons as this logic won't have been well tested until then.
                                                                            if (this.ubaBundleBlockRanges.length <= 30)
                                                                                return [2 /*return*/]; // TODO: Remove this line once we feel comfortable
                                                                            // with caching bundles.
                                                                            if (i === this.ubaBundleBlockRanges.length - 1)
                                                                                return [2 /*return*/];
                                                                            redisKeyForBundle = this.getKeyForBundle(mostRecentBundleBlockRanges, token, chainId);
                                                                            if (!isDefined(this.cachingMechanism)) return [3 /*break*/, 2];
                                                                            return [4 /*yield*/, this.cachingMechanism.set(redisKeyForBundle, modifiedFlowsInBundle[chainId]
                                                                                // I don't think we want these keys to expire since we'll likely always need data from the beginning
                                                                                // of the UBA activation block to validate even the latest bundles, because of the recursive nature
                                                                                // of how we compute running balances as a function of all prior validated bundle history.
                                                                                )];
                                                                        case 1:
                                                                            _a.sent();
                                                                            _a.label = 2;
                                                                        case 2: return [2 /*return*/];
                                                                    }
                                                                });
                                                            }); })];
                                                        case 4:
                                                            // Save into UBA client state
                                                            _a.sent();
                                                            return [2 /*return*/];
                                                    }
                                                });
                                            }); })];
                                    case 1:
                                        // console.log("Validating flows for bundle", mostRecentBundleBlockRanges);
                                        _c.sent();
                                        return [2 /*return*/];
                                }
                            });
                        };
                        this_4 = this;
                        i = 0;
                        _a.label = 5;
                    case 5:
                        if (!(i < this.ubaBundleBlockRanges.length)) return [3 /*break*/, 8];
                        return [5 /*yield**/, _loop_4(i)];
                    case 6:
                        _a.sent();
                        _a.label = 7;
                    case 7:
                        i++;
                        return [3 /*break*/, 5];
                    case 8:
                        _loop_5 = function (i) {
                            var mostRecentBundleBlockRanges = this_5.ubaBundleBlockRanges[i];
                            this_5.enabledChainIds.forEach(function (chainId) {
                                tokens.forEach(function (token) {
                                    var modifiedFlowsInBundle = _this.getBundleState(mostRecentBundleBlockRanges, token, chainId).flows;
                                    modifiedFlowsInBundle.forEach(function (_a) {
                                        var flow = _a.flow;
                                        // We can now set the realizedLpFeePct for the deposit in the SpokePoolClient, which was not
                                        // known to the SpokePoolClient at the time it queried the deposit.
                                        if (isUbaOutflow(flow) && !isDefined(flow.matchedDeposit.realizedLpFeePct)) {
                                            _this.spokePoolClients[flow.matchedDeposit.originChainId].updateDepositRealizedLpFeePct(flow.matchedDeposit, flow.realizedLpFeePct);
                                        }
                                    });
                                });
                            });
                        };
                        this_5 = this;
                        // For each validated outflow, update its matched deposit's realizedLpFeePct in the spoke pool client:
                        for (i = 0; i < this.ubaBundleBlockRanges.length; i++) {
                            _loop_5(i);
                        }
                        // At last, load into ubaClientState. This way the caller now can access bundle states two ways:
                        // via ubaCientState or via ubaBundleState.
                        this.ubaClientState = newUbaClientState;
                        this.isUpdated = true;
                        _loop_6 = function (i) {
                            tokens.forEach(function (token) {
                                var bundleBlockRange = _this.ubaBundleBlockRanges[i];
                                var breakdownPerChain = _this.enabledChainIds
                                    .map(function (chainId) {
                                    var bundleState = _this.getBundleState(bundleBlockRange, token, chainId);
                                    if (isDefined(bundleState)) {
                                        var flows = bundleState.flows;
                                        var lastFlow = flows.at(-1);
                                        if (!isDefined(lastFlow))
                                            return undefined;
                                        // Filter out outflows matched with pre UBA deposits from the list
                                        var fills = flows.filter(function (_a) {
                                            var flow = _a.flow;
                                            return isUbaOutflow(flow) && outflowIsFill(flow) && flow.matchedDeposit.realizedLpFeePct === undefined;
                                        });
                                        var deposits = flows.filter(function (_a) {
                                            var flow = _a.flow;
                                            return isUbaInflow(flow);
                                        });
                                        var refunds = flows.filter(function (_a) {
                                            var flow = _a.flow;
                                            return isUbaOutflow(flow) && outflowIsRefund(flow) && flow.matchedDeposit.realizedLpFeePct === undefined;
                                        });
                                        var inflows = deposits.reduce(function (sum, _a) {
                                            var flow = _a.flow;
                                            sum = sum.add(flow.amount);
                                            return sum;
                                        }, ethers.constants.Zero);
                                        var fillOutflows = fills.reduce(function (sum, _a) {
                                            var flow = _a.flow;
                                            sum = sum.add(flow.amount);
                                            return sum;
                                        }, ethers.constants.Zero);
                                        var refundOutflows = flows.reduce(function (sum, _a) {
                                            var balancingFee = _a.balancingFee;
                                            return sum.add(balancingFee);
                                        }, ethers.constants.Zero);
                                        var balancingFees = flows.reduce(function (sum, _a) {
                                            var balancingFee = _a.balancingFee;
                                            return sum.add(balancingFee);
                                        }, ethers.constants.Zero);
                                        var readableFlows = {
                                            fills: fills.length,
                                            fillRunningBalanceRemoved: fillOutflows.mul(-1).toString(),
                                            deposits: deposits.length,
                                            depositRunningBalanceAdded: inflows.toString(),
                                            refunds: refunds.length,
                                            refundRunningBalanceRemoved: refundOutflows.mul(-1).toString(),
                                            balancingFeesAdded: balancingFees.toString(),
                                            lpFeesAdded: flows
                                                .reduce(function (sum, _a) {
                                                var lpFee = _a.lpFee;
                                                return sum.add(lpFee);
                                            }, ethers.constants.Zero)
                                                .toString(),
                                            netRunningBalanceAdjustment: lastFlow.netRunningBalanceAdjustment.toString(),
                                            openingRunningBalance: bundleState.openingBalances.runningBalance.toString(),
                                            closingRunningBalance: lastFlow.runningBalance.toString(),
                                            openingIncentiveBalance: bundleState.openingBalances.incentiveBalance.toString(),
                                            closingIncentiveBalance: lastFlow.incentiveBalance.toString(),
                                        };
                                        // Sanity check:
                                        // - opening running balance minus outflows plus inflows minus balancing fees plus net running balance adjustments = closing running balance
                                        var expectedClosingBalance = bundleState.openingBalances.runningBalance
                                            .add(inflows)
                                            .sub(refundOutflows)
                                            .sub(fillOutflows)
                                            .sub(balancingFees)
                                            .add(lastFlow.netRunningBalanceAdjustment);
                                        if (!expectedClosingBalance.eq(lastFlow.runningBalance)) {
                                            _this.logger.error({
                                                at: "UBAClientWithRefresh#update",
                                                message: "Expected closing balance ".concat(expectedClosingBalance.toString(), " to equal actual closing balance ").concat(lastFlow.runningBalance.toString()),
                                                bundleBlockRange: bundleBlockRange,
                                                chainId: chainId,
                                                token: token,
                                                openingRunningBalance: bundleState.openingBalances.runningBalance.toString(),
                                                inflows: inflows.toString(),
                                                fillOutflows: fillOutflows.mul(-1).toString(),
                                                refundOutflows: refundOutflows.mul(-1).toString(),
                                                balancingFees: balancingFees.mul(-1).toString(),
                                                netRunningBalanceAdjustment: lastFlow.netRunningBalanceAdjustment.toString(),
                                            });
                                        }
                                        return [chainId, readableFlows];
                                    }
                                    else
                                        return undefined;
                                })
                                    .filter(isDefined);
                                var breakdown = Object.fromEntries(breakdownPerChain);
                                _this.logger.debug({
                                    at: "UBAClientWithRefresh#update",
                                    message: "Bundle state for ".concat(token),
                                    bundleBlockRange: bundleBlockRange,
                                    breakdown: breakdown,
                                });
                            });
                        };
                        // Log bundle states in readable form and check outputs:
                        for (i = 0; i < this.ubaBundleBlockRanges.length; i++) {
                            _loop_6(i);
                        }
                        return [2 /*return*/];
                }
            });
        });
    };
    return UBAClientWithRefresh;
}(BaseAbstractClient));
export { UBAClientWithRefresh };
//# sourceMappingURL=UBAClientWithRefresh.js.map