import { __awaiter, __generator } from "tslib";
import { getRelayHash } from "@across-protocol/contracts-v2/dist/test-utils";
/**
 * Find the block range that contains the deposit ID. This is a binary search that searches for the block range
 * that contains the deposit ID.
 * @param targetDepositId The target deposit ID to search for.
 * @param initLow The initial lower bound of the block range to search.
 * @param initHigh The initial upper bound of the block range to search.
 * @param maxSearches The maximum number of searches to perform. This is used to prevent infinite loops.
 * @returns The block range that contains the deposit ID.
 * @note  // We want to find the block range that satisfies these conditions:
 *        // - the low block has deposit count <= targetDepositId
 *        // - the high block has a deposit count > targetDepositId.
 *        // This way the caller can search for a FundsDeposited event between [low, high] that will always
 *        // contain the event emitted when deposit ID was incremented to targetDepositId + 1. This is the same transaction
 *        // where the deposit with deposit ID = targetDepositId was created.
 */
export function getBlockRangeForDepositId(targetDepositId, initLow, initHigh, maxSearches, spokePool) {
    return __awaiter(this, void 0, void 0, function () {
        var deploymentBlock, queriedIds, _getDepositIdAtBlock, _a, highestDepositIdInRange, lowestDepositIdInRange, low, high, searches, mid, midDepositId, accountedIdByMidBlock;
        var _this = this;
        return __generator(this, function (_b) {
            switch (_b.label) {
                case 0:
                    deploymentBlock = spokePool.deploymentBlock;
                    // Set the initial high block to the most recent block number or the initial high block, whichever is smaller.
                    initHigh = Math.min(initHigh, spokePool.latestBlockSearched);
                    // We will now set a list of sanity checks to ensure that the binary search will not fail
                    // due to invalid input parameters.
                    // If any of these sanity checks fail, then we will throw an error.
                    [
                        // Sanity check to ensure that the spoke pool client is updated
                        [spokePool.isUpdated, "Spoke pool client is not updated"],
                        // Sanity check to ensure that initHigh is greater than or equal to initLow.
                        [initLow <= initHigh, "Binary search failed because low > high"],
                        // Sanity check to ensure that init Low is greater than or equal to zero.
                        [initLow >= deploymentBlock, "Binary search failed because low must be >= deploymentBlock"],
                        // Sanity check to ensure that maxSearches is greater than zero.
                        [maxSearches > 0, "maxSearches must be > 0"],
                        // Sanity check to ensure that deploymentBlock is greater than or equal to zero.
                        [deploymentBlock >= 0, "deploymentBlock must be >= 0"],
                    ].forEach(function (_a) {
                        var condition = _a[0], errorMessage = _a[1];
                        // If the condition is false, then we will throw an error.
                        if (!condition) {
                            throw new Error(errorMessage);
                        }
                    });
                    queriedIds = {};
                    _getDepositIdAtBlock = function (blockNumber) { return __awaiter(_this, void 0, void 0, function () {
                        var _a, _b, _c;
                        var _d;
                        return __generator(this, function (_e) {
                            switch (_e.label) {
                                case 0:
                                    if (!((_d = queriedIds[blockNumber]) !== null && _d !== void 0)) return [3 /*break*/, 1];
                                    _a = _d;
                                    return [3 /*break*/, 3];
                                case 1:
                                    _b = queriedIds;
                                    _c = blockNumber;
                                    return [4 /*yield*/, spokePool._getDepositIdAtBlock(blockNumber)];
                                case 2:
                                    _a = (_b[_c] = _e.sent());
                                    _e.label = 3;
                                case 3:
                                    _a;
                                    return [2 /*return*/, queriedIds[blockNumber]];
                            }
                        });
                    }); };
                    return [4 /*yield*/, Promise.all([
                            _getDepositIdAtBlock(initHigh),
                            _getDepositIdAtBlock(Math.max(deploymentBlock, initLow - 1)),
                        ])];
                case 1:
                    _a = _b.sent(), highestDepositIdInRange = _a[0], lowestDepositIdInRange = _a[1];
                    // If the deposit ID at the initial high block is less than the target deposit ID, then we know that
                    // the target deposit ID must be greater than the initial high block, so we can throw an error.
                    if (highestDepositIdInRange <= targetDepositId) {
                        // initLow   = 5: Deposits Num: 10
                        //                                     // targetId = 11  <- fail (triggers this error)          // 10 <= 11
                        //                                     // targetId = 10  <- fail (triggers this error)          // 10 <= 10
                        //                                     // targetId = 09  <- pass (does not trigger this error)  // 10 <= 09
                        throw new Error("Target depositId is greater than the initial high block (".concat(targetDepositId, " > ").concat(highestDepositIdInRange, ")"));
                    }
                    // If the deposit ID at the initial low block is greater than the target deposit ID, then we know that
                    // the target deposit ID must be less than the initial low block, so we can throw an error.
                    if (lowestDepositIdInRange > targetDepositId) {
                        // initLow   = 5: Deposits Num: 10
                        // initLow-1 = 4: Deposits Num:  2
                        //                                     // targetId = 1 <- fail (triggers this error)
                        //                                     // targetId = 2 <- pass (does not trigger this error)
                        //                                     // targetId = 3 <- pass (does not trigger this error)
                        throw new Error("Target depositId is less than the initial low block (".concat(targetDepositId, " > ").concat(lowestDepositIdInRange, ")"));
                    }
                    low = initLow;
                    high = initHigh;
                    searches = 0;
                    _b.label = 2;
                case 2:
                    mid = Math.floor((low + high) / 2);
                    return [4 /*yield*/, _getDepositIdAtBlock(mid)];
                case 3:
                    midDepositId = _b.sent();
                    accountedIdByMidBlock = midDepositId - 1;
                    // If our target deposit ID is less than the smallest range of our
                    // midpoint deposit ID range, then we know that the target deposit ID
                    // must be in the lower half of the block range.
                    if (targetDepositId <= accountedIdByMidBlock) {
                        high = mid;
                    }
                    // If our target deposit ID is greater than the largest range of our
                    // midpoint deposit ID range, then we know that the target deposit ID
                    // must be in the upper half of the block range.
                    else {
                        low = mid + 1;
                    }
                    _b.label = 4;
                case 4:
                    if (++searches <= maxSearches && low < high) return [3 /*break*/, 2];
                    _b.label = 5;
                case 5:
                    // Sanity check to ensure that our low was not greater than our high.
                    if (low > high) {
                        throw new Error("Binary search failed (".concat(low, " > ").concat(high, "). SHOULD NEVER HAPPEN (but here we are)"));
                    }
                    // We've either found the block range or we've exceeded the maximum number of searches.
                    // In either case, the block range is [low, high] so we can return it.
                    return [2 /*return*/, { low: low, high: high }];
            }
        });
    });
}
/**
 * Finds the deposit id at a specific block number.
 * @param blockTag The block number to search for the deposit ID at.
 * @returns The deposit ID.
 */
export function getDepositIdAtBlock(contract, blockTag) {
    return __awaiter(this, void 0, void 0, function () {
        var depositIdAtBlock;
        return __generator(this, function (_a) {
            switch (_a.label) {
                case 0: return [4 /*yield*/, contract.numberOfDeposits({ blockTag: blockTag })];
                case 1:
                    depositIdAtBlock = _a.sent();
                    // Sanity check to ensure that the deposit ID is an integer and is greater than or equal to zero.
                    if (!Number.isInteger(depositIdAtBlock) || depositIdAtBlock < 0) {
                        throw new Error("Invalid deposit count");
                    }
                    return [2 /*return*/, depositIdAtBlock];
            }
        });
    });
}
export function relayFilledAmount(spokePool, relayData, blockTag) {
    var hash = getRelayHash(relayData.depositor, relayData.recipient, relayData.depositId, relayData.originChainId, relayData.destinationChainId, relayData.destinationToken, relayData.amount, relayData.realizedLpFeePct, relayData.relayerFeePct, relayData.message).relayHash;
    return spokePool.relayFills(hash, { blockTag: blockTag });
}
//# sourceMappingURL=SpokeUtils.js.map